# This file produces a calibrations of the whole typhoid model.
# Output files will be saved locally in the outputs folder.
# Desirable results must be manually moved to the calibrations folder so that they can be plotted by the
# Plot_final_typhoid_calibration.ipynb notebook. This is to avoid crowding the final results folder (calibrations) when
# trying out new ways of calibrating the model.


import atomica as at
import atomica.yaml_calibration
import os
import time
import shutil
import winsound

print(os.getcwd())
# run settings
yaml_dir = "../calibrations"
yaml_fname = "typ_calibration_instructions"
yaml_path = f"{yaml_dir}/{yaml_fname}.yaml"
testing = False  # <------------
savecals = True
dis_model = "typh"
out_path = "outputs"

# make project   ------------------------------------------------
print(os.getcwd())
inputs = "../assets"
F = at.ProjectFramework(f"{inputs}/T7_framework.xlsx")
D = at.ProjectData.from_spreadsheet(f"{inputs}/T7_databook.xlsx", framework=F)

P = at.Project(framework=F, databook=D, do_run=False)
P.settings.update_time_vector(start=2000, end=2050, dt=1 / 52)
cal = P.make_parset()
# starting calibration (if needed)
# cal.load_calibration(out_path + "/PAK_typh_YAML_autocalibrations_2024-04-25_1941_reinitialising/PAK_typh_YAML-autocalibration_2024-04-25_1959.xlsx")
# print('Existing calibration loaded')

# new folder to put calibrations in
date = time.strftime("%Y-%m-%d_%H%M")
calspath = f"{out_path}/cal_{yaml_fname}"
if testing:
    calspath += "_TESTING"
    savecals = False
if savecals:
    # save everything in folder
    try:
        os.makedirs(calspath)
    except FileExistsError:
        calspath += f"_{date}"
        os.makedirs(f"{calspath}")

    # save yaml file for reference
    new_fname = f"{yaml_fname}_{date}.yaml"
    shutil.copy(yaml_path, calspath)
    shutil.move(f"{calspath}/{yaml_fname}.yaml", f"{calspath}/{new_fname}.yaml")  # renaming file

    # save logfile
    at.start_logging(f"{calspath}/logfile_{date}.txt")

    # save current runfile for reference
    runfile_name = f"{os.path.basename(__file__)}_{date}"
    shutil.copy(__file__, f"{calspath}/{runfile_name}")

# display tree
cal_tree = at.yaml_calibration.build(yaml_path)
print(cal_tree)

# run calibration w yaml instructions # <--------------------------
newcal = P.calibrate(parset=cal, yaml=yaml_path, quiet=False, savedir=calspath, save_intermediate=False, verbose=0)

# save final calibration
date = time.strftime("%Y-%m-%d_%H%M")
newcal.save_calibration(f"{calspath}/typ_calibration.xlsx")
at.stop_logging()


# Notification - calibration finished
winsound.Beep(frequency=2500, duration=200)
winsound.Beep(frequency=2750, duration=200)
winsound.Beep(frequency=3050, duration=200)



# Simple modalities test
from atomica.programs import Covout
import numpy as np
from atomica import logger


def test_modalities():

    old_settings = np.seterr(all="raise")

    def run_test(coverage, outcomes, baseline, expected=None, imp_interaction=None):
        # cov - 1D vector of coverages
        # outcomes - 1D vector of outcomes
        # baseline - The baseline value
        # expected - target values for [additive, random, nested]
        if expected is None:
            expected = np.full((len(coverage),), np.nan)

        progs = ["P%d" % (x) for x in range(0, len(coverage))]
        coverage = {prog: [val] for prog, val in zip(progs, coverage)}

        covout = Covout(par="testpar", pop="testpop", cov_interaction="additive", imp_interaction=imp_interaction, baseline=baseline, progs={prog: val for prog, val in zip(progs, outcomes)})

        covout.cov_interaction = "additive"
        additive = covout.get_outcome(prop_covered=coverage)
        logger.debug("Additive = %.4f, Target = %.4f" % (additive, expected[0]))
        if np.isfinite(expected[0]):
            assert np.isclose(additive, expected[0], 1e-4)

        covout.cov_interaction = "random"
        random = covout.get_outcome(prop_covered=coverage)
        logger.debug("Random = %.4f, Target = %.4f" % (random, expected[1]))
        if np.isfinite(expected[1]):
            assert np.isclose(random, expected[1], 1e-4)

        covout.cov_interaction = "nested"
        nested = covout.get_outcome(prop_covered=coverage)
        logger.debug("Nested = %.4f, Target = %.4f" % (nested, expected[2]))
        if np.isfinite(expected[2]):
            assert np.isclose(nested, expected[2], 1e-4)

        return (additive, random, nested)  # Return outcomes

    # Test cases where we have an externally computed result
    # These should both run, and return a specific value
    run_test([0.25, 0.25, 0.25, 0.25], [0.8, 0.9, 0.4, 0.35], 0.3, [0.6125, 0.56308, 0.45])  # Coverages below 1.0
    run_test([0.25, 0.25, 0.25, 0.25], [0.8, 0.9, 0.4, 0.35], 0.3, [0.6125, 0.56308, 0.45])  # Coverages add to exactly 1.0
    run_test([0.2, 0.15, 0.3, 0.1], [50, 40, 20, 80], 5)  # Coverages all different, add to < 1
    run_test([0.244945, 0.022795, 0.056632], [0.8, 0.9, 0.4], 0.3, [0.4418126, 0.4375362, 0.4247518])  # 3 program test from HIV
    run_test([0.3, 0.2, 0.1, 0.05], [0.8, 0.9, 0.4, 0.35], 0.3, [0.5825, 0.54686, 0.47])
    run_test([1.0, 1.0, 1.0, 1.0], [0.8, 0.9, 0.4, 0.35], 0.3, [0.9, 0.9, 0.9])  # Tiebreaking case where all coverages are the same, and 1.0
    run_test([1.0, 1.0, 1.0, 1.0], [0.5, 0.5, 0.5, 0.5], 0.3, [0.5, 0.5, 0.5])  # Tiebreaking case where all outcomes and coverages are the same
    run_test([0.1, 0.0, 0.2, 0.3], [0.5, 0.5, 0.5, 0.5], 0.3, [0.42, 0.3992, 0.36])  # One of them has zero coverage, total coverage < 1
    run_test([0.6, 0.0, 0.2, 0.3], [0.5, 0.5, 0.5, 0.5], 0.3, [0.5, 0.4552, 0.42])  # One of them has zero coverage, total coverage > 1
    run_test([0.0, 0.0, 0.0, 0.0], [0.5, 0.5, 0.5, 0.5], 0.3, [0.3, 0.3, 0.3])  # All of them have zero coverage
    run_test([0.3, 0.3, 0.3, 0.3], [0.8, 0.9, 0.4, 0.35], 0.3, [0.665, 0.604845, 0.48])  # Coverages all same, and sum to > 1

    # Different, duplicate,identical,includes 1
    # - Coverages add to <        [0.05,0.1,0.15,0.2],[0.05,0.15,0.15,0.2],[0.2,0.2,0.2,0.2]
    # - Coverages exactly 1       [0.1,0.2,0.3,0.4]  ,[0.1,0.2,0.2,0.5] , [0.3,0.3,0.3,0.3], [0,0,0,1]
    # - Coverages > 1             [0.2,0.3,0.4,0.5]  ,[0.1,0.15,0.2,0.25], [0.4,0.4,0.4,0.4], [1,1,1,1]

    # Next, check exhaustively for numerical edge cases. There are too many combinations to check explicitly
    # so just run them to check that an error does not occur
    coverages = [[0.05, 0.1, 0.15, 0.2], [0.05, 0.15, 0.15, 0.2], [0.2, 0.2, 0.2, 0.2], [0.1, 0.2, 0.3, 0.4], [0.1, 0.2, 0.2, 0.5], [0.3, 0.3, 0.3, 0.3], [0.0, 0.0, 0.0, 1.0], [0.2, 0.3, 0.4, 0.5], [0.1, 0.15, 0.2, 0.25], [0.4, 0.4, 0.4, 0.4], [1.0, 1.0, 1.0, 1.0]]
    outcomes = [[0.05, 0.1, 0.15, 0.2], [0.05, 0.15, 0.15, 0.2], [0.2, 0.2, 0.2, 0.2], [0.1, 0.2, 0.3, 0.4], [0.1, 0.2, 0.2, 0.5], [0.3, 0.3, 0.3, 0.3], [0.0, 0.0, 0.0, 1.0], [0.2, 0.3, 0.4, 0.5], [0.1, 0.15, 0.2, 0.25], [0.4, 0.4, 0.4, 0.4], [1.0, 1.0, 1.0, 1.0]]
    baselines = [-0.1, 0, 0.1, 0.2]
    for coverage in coverages:
        for outcome in outcomes:
            for baseline in baselines:
                run_test(coverage, outcome, baseline)

    # Check some of the expected properties
    base = run_test([0.4, 0.3, 0.2, 0.1], [0.8, 0.9, 0.4, 0.35], 0.3, [0.705, 0.63008, 0.53])
    improved = run_test([0.4, 0.31, 0.2, 0.1], [0.8, 0.9, 0.4, 0.35], 0.3, [0.7105, 0.633936, 0.531])  # Crossing the threshold increases coverage (whereas it decreased it before)
    assert base[0] < improved[0]

    # Test effect of overriding the interaction
    base = run_test([0.25, 0.25, 0.25, 0.25], [0.8, 0.9, 0.4, 0.35], 0.3)  # Baseline result
    better = run_test([0.25, 0.25, 0.25, 0.25], [0.8, 0.9, 0.4, 0.35], 0.3, imp_interaction="P1+P2=0.95")
    assert better[1] > base[1]
    worse = run_test([0.25, 0.25, 0.25, 0.25], [0.8, 0.9, 0.4, 0.35], 0.3, imp_interaction="P1+P2=0.85")
    assert worse[1] < base[1]
    same = run_test([0.25, 0.25, 0.25, 0.25], [0.8, 0.9, 0.4, 0.35], 0.3, imp_interaction="P1+P2=0.90")
    assert same[1] == base[1]
    better = run_test([0.25, 0.25, 0.25, 0.25], [0.8, 0.9, 0.4, 0.35], 0.3, imp_interaction="P1+P2=0.9,P1+P2+P3=1.0")
    assert better[1] > base[1]
    better = run_test([0.25, 0.25, 0.25, 0.25], [0.8, 0.9, 0.4, 0.35], 0.3, imp_interaction="P0+P1+P2+P3=1.0")
    assert better[2] == (0.25 * 1.0 + 0.75 * 0.3)  # Check that the nested interaction matches known result

    np.seterr(**old_settings)  # Reset numpy error behaviour


if __name__ == "__main__":
    test_modalities()
    print("All tests completed successfully")



import atomica as at
import numpy as np
import os

testdir = at.parent_dir()  # Must be relative to current file to work with tox


def test_indirect_programs():

    F = at.ProjectFramework(testdir / "test_indirect_programs_framework.xlsx")
    D = at.ProjectData.new(F, tvec=[2019], pops=1, transfers=0)

    P = at.Project(framework=F, databook=D, do_run=False)
    P.settings.update_time_vector(start=2019, end=2025, dt=1)
    res = P.run_sim()

    np.testing.assert_array_equal(res.get_variable("par_1")[0].vals, res.get_variable("v_rate")[0].vals)
    np.testing.assert_array_equal(res.get_variable("par_1")[0].vals, res.get_variable("par_2")[0].vals)

    P.load_progbook(testdir / "test_indirect_programs_progbook.xlsx")
    ins = at.ProgramInstructions(start_year=2019)
    res = P.run_sim(parset=0, progset=0, progset_instructions=ins)

    np.testing.assert_array_equal(res.get_variable("par_1")[0].vals, res.get_variable("v_rate")[0].vals)
    np.testing.assert_array_equal(res.get_variable("par_1")[0].vals, res.get_variable("par_2")[0].vals)


if __name__ == "__main__":
    test_indirect_programs()



# This script tests some databook IO operations

import atomica as at
from atomica.excel import transfer_comments
import numpy as np
from atomica import ProjectFramework, Project, ProjectData
import sciris as sc

testdir = at.parent_dir()
tmpdir = testdir / "temp"


def test_databooks():

    F = ProjectFramework(at.LIBRARY_PATH / "tb_framework.xlsx")
    F.save(tmpdir / "f_blug.xlsx")

    # Copy a databook by loading and saving it
    data = ProjectData.from_spreadsheet(at.LIBRARY_PATH / "tb_databook.xlsx", F)
    data.save(tmpdir / "d_blug.xlsx")

    # Copy comments, using lower-level Spreadsheet object (for in-memory file operations)
    original_workbook = sc.Spreadsheet(at.LIBRARY_PATH / "tb_databook.xlsx")
    new_workbook = data.to_spreadsheet()  # This is a sc.Spreadsheet that can be stored in the FE database
    transfer_comments(new_workbook, original_workbook)
    new_workbook.save(tmpdir / "d_blug_formatted.xlsx")

    # Run the copied databook
    P = Project(name="test", framework=F, do_run=False)
    P.load_databook(databook_path=tmpdir / "d_blug.xlsx", make_default_parset=True, do_run=True)
    d = at.PlotData(P.results["parset_default"], pops="0-4")
    at.plot_series(d, plot_type="stacked")  # This should look like the usual Optima-TB result

    # Change the time axis
    d2 = sc.dcp(data)
    d2.change_tvec(np.arange(2000, 2017, 0.5))
    d2.transfers[0].comment = "Test comment"
    d2.save(tmpdir / "d_blug_halfyear.xlsx")

    # Run the half-year databook
    P = Project(name="test", framework=F, do_run=False)
    P.load_databook(databook_path=tmpdir / "d_blug_halfyear.xlsx", make_default_parset=True, do_run=True)
    d = at.PlotData(P.results["parset_default"], pops="0-4")
    at.plot_series(d, plot_type="stacked")  # This should look like the usual Optima-TB result

    # Change the pops and run it
    data = ProjectData.from_spreadsheet(at.LIBRARY_PATH / "tb_databook.xlsx", F)
    data.rename_pop("0-4", "0-3", "Gen 0-3")
    data.rename_transfer("age", "age_up", "Age Up")
    data.save(tmpdir / "d_blug_renamed.xlsx")
    P = Project(name="test", framework=F, do_run=False)
    P.load_databook(databook_path=tmpdir / "d_blug_renamed.xlsx", make_default_parset=True, do_run=True)
    d = at.PlotData(P.results["parset_default"], pops="0-3")
    at.plot_series(d, plot_type="stacked")  # This should look like the usual Optima-TB result

    # Remove a key pop
    d2 = sc.dcp(data)
    d2.remove_pop("Prisoners")
    d2.save(tmpdir / "d_blug_nopris.xlsx")

    # Remove a transfer, add an interaction, add a pop
    d2.remove_transfer("inc")
    d2.add_interaction("d_ctc", "New interpop")
    d2.add_pop("asdf", "The ASDF pop")
    d2.save(tmpdir / "d_blug_newpop.xlsx")

    # Make a brand new databook
    d2 = ProjectData.new(F, np.arange(2000, 2017), pops=2, transfers=4)
    d2.save(tmpdir / "d_blug_blank.xlsx")

    # Make a blank databook with the same pops and transfers from the old one
    pops = sc.odict()
    for pop, val in data.pops.items():
        pops[pop] = val["label"]

    transfers = sc.odict()
    for transfer in data.transfers:
        transfers[transfer.code_name] = transfer.full_name

    d2 = ProjectData.new(F, np.arange(2000, 2017), pops=pops, transfers=transfers)
    d2.save(tmpdir / "d_cleared.xlsx")

    # Modify incomplete databook
    d2 = ProjectData.from_spreadsheet(tmpdir / "d_blug_blank.xlsx", F)
    d2.add_pop("asdf", "The ASDF pop")
    d2.add_interaction("d_ctc", "New interpop")
    d2.save(tmpdir / "d_blug_blank_modified.xlsx")

    # Test writing out a databook with uncertainty values
    data = ProjectData.from_spreadsheet(at.LIBRARY_PATH / "tb_databook.xlsx", F)
    data.tdve["alive"].ts[0].sigma = 100

    # Check stripping uncertainty values
    for tdve in data.tdve.values():
        tdve.write_uncertainty = True  # Force writing uncertainty
    data.save(tmpdir / "d_blug_uncertainty.xlsx")  # Every table has uncertainty - this is the default

    # Check stripping uncertainty values
    for tdve in data.tdve.values():
        tdve.write_uncertainty = False
    data.save(tmpdir / "d_blug_uncertainty_stripped.xlsx")  # No tables have uncertainty

    # Check the values get read in correctly
    d3 = at.ProjectData.from_spreadsheet(tmpdir / "d_blug_uncertainty.xlsx", F)
    assert d3.tdve["alive"].ts[0].sigma == 100
    d3 = at.ProjectData.from_spreadsheet(tmpdir / "d_blug_uncertainty_stripped.xlsx", F)
    assert d3.tdve["alive"].ts[0].sigma is None


def test_mixed_years_1():
    F = ProjectFramework(at.LIBRARY_PATH / "sir_framework.xlsx")

    D = at.ProjectData.from_spreadsheet(testdir / "test_databook_mixed_years_1.xlsx", framework=F)
    assert 2020 in D.tvec  # This should get picked up from the second table
    D.add_interaction("d_ctc", "New interpop")
    D.save(tmpdir / "test_databook_mixed_years_1.xlsx")


def test_mixed_years_2():
    F = ProjectFramework(testdir / "test_no_compartment_framework.xlsx")
    D = at.ProjectData.from_spreadsheet(testdir / "test_no_compartment_databook_mixed.xlsx", framework=F)
    assert len(D.tvec) == 0
    D.add_interaction("d_ctc", "New interpop")
    D.save(tmpdir / "test_databook_mixed_years_2a.xlsx")
    D = at.ProjectData.from_spreadsheet(tmpdir / "test_databook_mixed_years_2a.xlsx", framework=F)
    assert len(D.tvec) == 0

    D = at.ProjectData.from_spreadsheet(testdir / "test_no_compartment_databook.xlsx", framework=F)
    D.add_interaction("d_ctc", "New interpop")
    D.save(tmpdir / "test_databook_mixed_years_2b.xlsx")
    D = at.ProjectData.from_spreadsheet(tmpdir / "test_databook_mixed_years_2b.xlsx", framework=F)
    assert len(D.tvec) == 1


def test_databook_comments():
    F = ProjectFramework(at.LIBRARY_PATH / "sir_framework.xlsx")
    D = at.ProjectData.from_spreadsheet(testdir / "sir_databook_comment_test.xlsx", framework=F)
    assert D.get_ts("sus", "Children 0-4").get(2000) == 1  # Should not overwrite existing value
    assert D.get_ts("age", ("Children 0-4", "Children 5-14")).get(2004) == 14  # Should not overwrite existing value
    assert D.get_ts("transfer_1", ("Children 0-4", "Adults 15-64")).get(2000) == 10  # Should not overwrite existing value
    assert D.get_ts("recrate", "Children 0-4").get(2000) == 0.5
    assert D.get_ts("infdeath", "Children 0-4").get(2000) == 0.6
    assert D.transfers[0].attributes["Extra attribute"] == "value"


def test_databook_all():

    F = ProjectFramework(at.LIBRARY_PATH / "sir_framework.xlsx")
    # D = at.ProjectData.new(framework=F, tvec=np.arange(2020,2023),pops=4,transfers=1)
    # D.save(testdir / "sir_databook_all.xlsx")
    D = at.ProjectData.from_spreadsheet(testdir / "sir_databook_all.xlsx", framework=F)

    P = at.Project(framework=F, databook=D, do_run=False)
    P.data.save(tmpdir / "databook_all_test.xlsx")  # Test saving it back
    res = P.run_sim()

    assert len(P.data.tdve["sus"].ts) == 1
    assert len(P.parsets[0].get_par("sus").ts) == 4
    assert len(P.parsets[0].get_par("sus").y_factor) == 4

    # Check all populations have the same susceptible
    assert np.isclose(res.get_variable("sus", "pop_0")[0].vals[0], 100)
    assert np.isclose(res.get_variable("sus", "pop_1")[0].vals[0], 100)
    assert np.isclose(res.get_variable("sus", "pop_2")[0].vals[0], 100)
    assert np.isclose(res.get_variable("sus", "pop_3")[0].vals[0], 100)

    # Check prevalence calculation - 50% prevalence in all populations
    # and 100 susceptible, with pop sizes of 200, 300, 400, 500 should give
    # different values for the inf and rec compartment sizes
    assert np.isclose(res.get_variable("inf", "pop_0")[0].vals[0], 100)
    assert np.isclose(res.get_variable("inf", "pop_1")[0].vals[0], 30)
    assert np.isclose(res.get_variable("inf", "pop_2")[0].vals[0], 200)
    assert np.isclose(res.get_variable("inf", "pop_3")[0].vals[0], 250)

    assert np.isclose(res.get_variable("rec", "pop_0")[0].vals[0], 0)
    assert np.isclose(res.get_variable("rec", "pop_1")[0].vals[0], 170)
    assert np.isclose(res.get_variable("rec", "pop_2")[0].vals[0], 100)
    assert np.isclose(res.get_variable("rec", "pop_3")[0].vals[0], 150)

    d = at.PlotData(res, outputs=["sus", "ch_prev", "ch_all"], project=P)
    at.plot_series(d, axis="pops", data=P.data)


def test_databook_default_all():
    F = at.ProjectFramework(testdir / "sir_framework_default_all.xlsx")
    D = at.ProjectData.new(F, np.arange(2000, 2005), pops=4, transfers=1)
    D.save(tmpdir / "databook_default_all_test.xlsx")  # Test saving it back

    D.add_pop("new", "new")
    assert list(D.tdve["sus"].ts.keys()) == ["pop_0", "pop_1", "pop_2", "pop_3", "new"]
    assert list(D.tdve["ch_prev"].ts.keys()) == ["All"]

    D.tdve["contacts"].ts["pop_0"] = D.tdve["contacts"].ts["All"]
    del D.tdve["contacts"].ts["All"]
    D.add_pop("new2", "new2")
    assert list(D.tdve["ch_prev"].ts.keys()) == ["All"]
    assert list(D.tdve["contacts"].ts.keys()) == ["pop_0", "new2"]
    D.save(tmpdir / "databook_default_all_test_2.xlsx")  # Test saving it back


if __name__ == "__main__":
    # test_mixed_years_2()
    # test_mixed_years_1()
    # test_databooks()
    # test_databook_comments()
    # test_databook_all()
    test_databook_default_all()



# Test time vector
# The problem:
# In atomica, we want to include the end year: Say we're running a simulation from 2000 to 2050 ->
# we want to be able to get the results for year 2000, 2001, ... and 2050. Hence, we need to run
# more time steps after 2050, in particular up to 2051-1/12

# Currently, at.ProjectSettings.tvec creates 611 entries -> if taken the integer,the last year doesn't have enough entries

import atomica as at
import numpy as np
import pandas as pd


def test_time_vector():
    # Test 1
    # Creating a setting object
    analysis_years = (2000, 2050)
    dt = 1 / 12
    settings = at.ProjectSettings(sim_start=analysis_years[0], sim_end=analysis_years[1] + 1 - dt, sim_dt=dt)  #: Atomica project settings
    tvec = settings.tvec  #: Simulation/result time points
    assert set(np.unique(tvec.astype(int), return_counts=True)[1]) == {1 / dt}  # or pd.Series(tvec).astype(int).value_counts()


if __name__ == "__main__":
    test_time_vector()



# This test checks that the error handling code is OK
# One problem with only testing valid inputs is that typos in error handling code
# (e.g. badly formatted messages) only happen in production. Thus we just test a subset
# of bad inputs here to make sure the correct error is raised

import atomica as at
import pytest

testdir = at.parent_dir()  # Must be relative to current file to work with tox
fdir = testdir / "bad_frameworks"

# List available models based on which framework files exist
framework_files = list()
for f in fdir.iterdir():
    if f.name.endswith(".xlsx") and not f.name.startswith("~$"):
        framework_files.append(f)


@pytest.mark.parametrize("framework_file", framework_files)
def test_bad_framework(framework_file):
    with pytest.raises(at.InvalidFramework):
        at.ProjectFramework(framework_file)


if __name__ == "__main__":
    for framework_file in framework_files:
        print("Checking %s" % (framework_file))
        test_bad_framework(framework_file)



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

import matplotlib

matplotlib.use("agg")

import os
import sys

sys.path.insert(0, os.path.abspath("../"))  # Source code dir relative to this file

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# -- Project information -----------------------------------------------------

project = "Atomica"
copyright = "2025, Atomica Team"
author = "Atomica Team"

import atomica

# The short X.Y version
version = atomica.__version__
# The full version, including alpha/beta/rc tags
release = ""


# -- General configuration ---------------------------------------------------

os.environ["PYDEVD_DISABLE_FILE_VALIDATION"] = "1"  # Suppress harmless warning in documentation build

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",  # Core Sphinx library for auto html doc generation from docstrings
    "sphinx.ext.autosummary",  # Create neat summary tables for modules/classes/methods etc
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",  # Add a link to the Python source code for classes, functions etc.
    "sphinx.ext.napoleon",
    "sphinx.ext.autosectionlabel",
    "sphinx_autodoc_typehints",  # Automatically document param types (less noise in class signature)
    "sphinx_markdown_tables",
    "recommonmark",
    "nbsphinx",
    "sphinx.ext.intersphinx",  # Link to other project's documentation (see mapping below)
    "IPython.sphinxext.ipython_console_highlighting",  # Temporary fix for https://github.com/spatialaudio/nbsphinx/issues/687
    "sphinxcontrib.jquery",  # Fix for https://github.com/readthedocs/sphinx_rtd_theme/issues/1452
]

# Configure intersphinx
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sciris": ("https://sciris.readthedocs.io/en/latest/", None),  # This doesn't really work though, because things are accessed via sciris.Spreadsheet not sciris.sc_file.Spreadshee
}

# Configure autosummary
autosummary_generate = True  # Turn on sphinx.ext.autosummary
autoclass_content = "both"  # Add __init__ doc (ie. params) to class summaries
html_show_sourcelink = False  # Remove 'view source code' from top of page (for html, not python)
# autodoc_default_options = {
#     'members': False,
#     'private-members': False,
#     'undoc-members': False,
#     'ignore-module-all': False,
# }

# Configure nbsphinx
nbsphinx_kernel_name = "python"
nbsphinx_timeout = -1  # Disable timeout for slow cells
nbsphinx_execute_arguments = [
    "--InlineBackend.figure_formats={'svg', 'pdf'}",
    "--InlineBackend.rc=figure.dpi=96",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "**.ipynb_checkpoints",
    "general/programs/saturation.ipynb",  # This is not a documentation page - just for generating a figure
]


# -- Options for HTML output -------------------------------------------------

# Use RTD
import sphinx_rtd_theme

html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
html_theme_options = {
    "collapse_navigation": False,
    "navigation_depth": -1,
}



# Test safe division in function_parser

from atomica.function_parser import sdiv
import numpy as np


def test_sdiv():

    with np.errstate(divide="ignore"):  # Suppress expected numpy division by zero warning

        # Both scalar
        assert sdiv(1, 2) == 0.5
        assert sdiv(1, 0) == np.inf
        assert sdiv(0, 1) == 0
        assert sdiv(0, 0) == 0

        # Numerator vector
        assert np.allclose(sdiv(np.array(np.array([0, 1])), 0), np.array([0, np.inf]), equal_nan=True)

        # Denominator vector
        assert np.allclose(sdiv(0, np.array([0, 1])), np.array([0, 0]), equal_nan=True)

        # Both vector
        assert np.allclose(sdiv(np.array([0, 0, 1, 1]), np.array([0, 1, 0, 1])), np.array([0, 0, np.inf, 1]), equal_nan=True)

        # Different shape
        assert np.allclose(sdiv(np.array([[0, 0], [1, 1]]), np.array([[0, 1], [0, 1]])), np.array([[0, 0], [np.inf, 1]]), equal_nan=True)


if __name__ == "__main__":
    test_sdiv()



# This script performs various tests on Result objects

import numpy as np
import atomica as at
import matplotlib.pyplot as plt
import os
import sciris as sc
import pytest


testdir = at.parent_dir()
tmpdir = testdir / "temp"


def test_export():
    P = at.demo("tb", do_run=False)
    P.run_sim(parset="default", result_name="parset1", store_results=True)
    P.run_sim(parset="default", result_name="parset2", store_results=True)
    instructions = at.ProgramInstructions(start_year=2018)
    P.run_sim(parset="default", progset="default", progset_instructions=instructions, result_name="progset1", store_results=True)
    P.run_sim(parset="default", progset="default", progset_instructions=instructions, result_name="progset2", store_results=True)

    # Test export single
    at.export_results(P.results["parset1"], tmpdir / "export_parset.xlsx")
    at.export_results(P.results["progset1"], tmpdir / "export_progset.xlsx")

    # Test export multi
    at.export_results(P.results, tmpdir / "export_multi.xlsx")
    output_ordering = ("pop", "output", "result")
    cascade_ordering = ("pop", "stage", "result")
    program_ordering = ("quantity", "program", "result")
    at.export_results(P.results, tmpdir / "export_multi_reordered.xlsx", output_ordering=output_ordering, cascade_ordering=cascade_ordering, program_ordering=program_ordering)

    # Test raw exports
    P.results["parset1"].export_raw(tmpdir / "export_raw_parset.xlsx")
    P.results["progset1"].export_raw(tmpdir / "export_raw_progset.xlsx")


if __name__ == "__main__":
    test_export()



# Run various framework-related tests

import atomica as at
import numpy as np
import pytest
import pandas as pd

testdir = at.parent_dir()
tmpdir = testdir / "temp"

frameworks = [
    "framework_blank_sheet.xlsx",
    "framework_mixed_pages.xlsx",
    "framework_no_pages.xlsx",
    "framework_extra_pages.xlsx",
    "framework_missing_page_column.xlsx",
]


@pytest.mark.parametrize("fname", frameworks)
def test_basic_framework(fname):
    F = at.ProjectFramework(testdir / fname)
    D = at.ProjectData.new(F, np.arange(2020, 2022), 1, 1)
    D.save(tmpdir / fname.replace("framework", "databook"))


def test_framework_par_min_max():
    # Check that the vector min/max functions work properly

    F_path = testdir / "framework_par_min_max_test.xlsx"
    F = at.ProjectFramework(F_path)

    # D = at.ProjectData.new(F,[2020,2021,2022],1,0)
    # D.save(testdir / 'par_min_max_databook.xlsx')
    D = at.ProjectData.from_spreadsheet(testdir / "par_min_max_databook.xlsx", F)

    # Example of running a bare model without a project
    res = at.run_model(at.ProjectSettings(2020, 2022, 1), F, at.ParameterSet(F, D))

    minpar = res.get_variable("minpar")[0].vals
    maxpar = res.get_variable("maxpar")[0].vals

    assert np.array_equal(minpar, [2, 5, 4])
    assert np.array_equal(maxpar, [5, 10, 6])


def test_framework_single_char():
    # Check that single chracter variable names work

    P1 = at.demo("sir")
    r1 = P1.results[0]

    F = at.ProjectFramework(testdir / "test_single_char_framework.xlsx")
    D = at.ProjectData.new(tvec=P1.data.tvec, framework=F, pops=P1.data.pops, transfers=0)
    P = at.Project(framework=F, databook=D, do_run=False)
    P.settings = P1.settings
    r2 = P.run_sim()

    test_equal = lambda var1, var2: np.testing.assert_array_equal(r1.get_variable(var1)[0].vals, r2.get_variable(var2)[0].vals)

    test_equal("sus", "s")
    test_equal("inf", "i")
    test_equal("rec", "r")


def test_framework_spaces():
    F = at.ProjectFramework(testdir / "test_framework_spaces.xlsx")
    F.get_variable("Transmission probability per contact")  # Has no leading or trailing spaces in the framework
    F.get_variable("Number of contacts annually")  # Has a leading space in the framework
    F.get_variable("Death rate for infected people")  # Has a trailing space in the framework


def test_framework_comments():
    F1 = at.ProjectFramework(testdir / "sir_framework_ignore.xlsx")
    F2 = at.ProjectFramework(at.LIBRARY_PATH / "sir_framework.xlsx")

    assert F1.transitions == F2.transitions
    assert F1.pars.equals(F2.pars)
    assert F1.characs.at["ch_prev", "default value"] == 10
    assert pd.isna(F1.characs.at["ch_propnewinf", "default value"])  # This will not be NaN if the #ignore was ignored


if __name__ == "__main__":

    for fname in frameworks:
        test_basic_framework(fname)

    test_framework_par_min_max()
    test_framework_single_char()
    test_framework_spaces()



# This script checks that the junction update is correct

import numpy as np
import atomica as at
import os

testdir = at.parent_dir()  # Must be relative to current file to work with tox


def test_junctions():

    F = at.ProjectFramework(testdir / "framework_junction_test.xlsx")
    D = at.ProjectData.new(F, np.arange(2000, 2001), pops={"pop1": "Population 1"}, transfers=0)

    P = at.Project(name="test", framework=F, do_run=False)
    P.settings.update_time_vector(dt=0.25)
    P.load_databook(databook_path=D.to_spreadsheet(), make_default_parset=True, do_run=True)

    d = at.PlotData(P.results[0], [":ca", ":cb", "cb:", ":cc", ":ce", "ce"], project=P)
    at.plot_series(d, axis="pops")

    # Do some validation checks
    pop = P.results[0].model.pops[0]

    # The junction setup is this - B, C, and D are all junctions. We have
    # B -> C,D (50/50)
    # C -> E,F (50/50)
    # D -> G,H (50/50)
    #
    # We also have that junction C has an extra 100 people/year after the simulation starts

    # CHECK INITIAL FLUSH
    # Initially, the junctions are all initialized as having 500 people. And there is a 50% split between
    # all junctions. Therefore, we should have
    # Junction C and D acquire 250 people at the first timestep, for a total of 750 people
    # Compartments E, F, G, H each get 750/2=375 people at the first timestep

    # Check junctions are flushed
    assert pop.get_comp("cb").vals[0] == 0
    assert pop.get_comp("cc").vals[0] == 0
    assert pop.get_comp("cd").vals[0] == 0

    assert pop.get_comp("ce").vals[0] == 375
    assert pop.get_comp("cf").vals[0] == 375
    assert pop.get_comp("cg").vals[0] == 375
    assert pop.get_comp("ch").vals[0] == 375

    # After this initial flush, the junctions should now be empty. However, junction C has an inflow of
    # 100 people per year directly from the input. Therefore, the flow into E and F should be 25/2=12.5, while
    # flow into G and H should be zero
    assert pop.get_links("jb1:flow")[0].vals[0] == 0
    assert pop.get_links("jb2:flow")[0].vals[0] == 0
    assert pop.get_links("jc1:flow")[0].vals[0] == 12.5
    assert pop.get_links("jc2:flow")[0].vals[0] == 12.5
    assert pop.get_links("jd1:flow")[0].vals[0] == 0
    assert pop.get_links("jd2:flow")[0].vals[0] == 0

    # Then, after the first timestep, there are 25 people in compartment A
    assert pop.get_comp("ca").vals[1] == 25

    # Which means that there are scheduled to be 25/4 people removed from A
    # (i.e. an annual probability of 1 means that 25% of the compartment is removed this timestep)
    assert pop.get_links("a1")[0].vals[1] == 25 / 4

    # Which means that there should have been be 50% of this number transferred to C and D via those links
    # at that timestep
    assert pop.get_links("jb1")[0].vals[1] == 25 / 4 / 2
    assert pop.get_links("jb2")[0].vals[1] == 25 / 4 / 2

    # Then, junction C has a total inflow of 3.125 people from B, and 25 people from src. Thus
    # the flow to E and F should be (3.125+25)/2=14.0625, while the flow to G and H should be
    # 3.125/2=1.5625. And that should be the same at ALL subsequent times (no NaN at the end)
    assert np.all(pop.get_links("jc1:flow")[0].vals[1] == 14.0625)
    assert np.all(pop.get_links("jc2:flow")[0].vals[1] == 14.0625)
    assert np.all(pop.get_links("jd1:flow")[0].vals[1] == 1.5625)
    assert np.all(pop.get_links("jd2:flow")[0].vals[1] == 1.5625)

    # The equilibrium solution has 100 people/year flowing into A with 100% outflow implying
    # 25 people flowing out per timestep. Thus the final values should be that 25 people get
    # removed per timestep from A, not 25/4. So in the long term, we converge asymptotically
    # to the values below - no NaN at the end
    assert np.allclose(pop.get_links("jc1:flow")[0].vals[-1], 18.75, 1e-3)
    assert np.allclose(pop.get_links("jc2:flow")[0].vals[-1], 18.75, 1e-3)
    assert np.allclose(pop.get_links("jd1:flow")[0].vals[-1], 6.25, 1e-3)
    assert np.allclose(pop.get_links("jd2:flow")[0].vals[-1], 6.25, 1e-3)

    print("Test successfully completed")


def test_only_junctions():
    # This is a minimal framework that has a source connected only to junctions and then sinks
    # So there are no normal compartments
    F = at.ProjectFramework(testdir / "test_only_junctions_framework.xlsx")
    D = at.ProjectData.new(F, [2018], pops=1, transfers=0)

    P = at.Project(name="test", framework=F, databook=D.to_spreadsheet(), do_run=True)
    res = P.results[0]

    assert res.get_variable("state1")[0].vals[0] == 0
    assert res.get_variable("state1")[0].vals[1] == 40
    assert res.get_variable("state1")[0].vals[2] == 80

    assert res.get_variable("state2")[0].vals[0] == 0
    assert res.get_variable("state2")[0].vals[1] == 60
    assert res.get_variable("state2")[0].vals[2] == 120


def test_junction_remainder():
    F = at.ProjectFramework(testdir / "framework_junction_remainder_test.xlsx")
    D = at.ProjectData.new(F, [2018], pops=1, transfers=0)

    P = at.Project(name="test", framework=F, databook=D.to_spreadsheet(), do_run=True)
    res = P.results[0]

    # Initial flush
    assert res.get_variable("s1")[0].vals[0] == 60
    assert res.get_variable("s2")[0].vals[0] == 30

    # First timestep
    assert res.get_variable("s1")[0].vals[1] == 60 + 0.6 * 200 * P.settings.sim_dt
    assert res.get_variable("s2")[0].vals[1] == 30 + 0.3 * 200 * P.settings.sim_dt
    assert res.get_variable("j1:j2")[0].vals[1] == 0.1 * 200 * P.settings.sim_dt


def test_junction_remainder_2():
    F = at.ProjectFramework(testdir / "framework_junction_remainder_test_2.xlsx")
    D = at.ProjectData.new(F, [2018], pops=1, transfers=0)

    P = at.Project(name="test", framework=F, databook=D.to_spreadsheet(), do_run=True)
    res = P.results[0]

    # Initial flush
    assert res.get_variable("a1")[0].vals[0] == 20
    assert res.get_variable("b1")[0].vals[0] == 60
    assert res.get_variable("c1")[0].vals[0] == 20

    assert res.get_variable("a2")[0].vals[0] == 10
    assert res.get_variable("b2")[0].vals[0] == 30
    assert res.get_variable("c2")[0].vals[0] == 60

    assert res.get_variable("a3")[0].vals[0] == 100 / 3
    assert res.get_variable("b3")[0].vals[0] == 100 / 3
    assert res.get_variable("c3")[0].vals[0] == 100 / 3

    assert res.get_variable("a4")[0].vals[0] == 50
    assert res.get_variable("b4")[0].vals[0] == 50
    assert res.get_variable("c4")[0].vals[0] == 0


def test_junction_timed_remainder():
    """
    Test a sequence of junctions with a duration group
    """

    F = at.ProjectFramework(testdir / "framework_junction_timed_remainder_test.xlsx")
    D = at.ProjectData.new(F, [2018], pops=1, transfers=0)
    P = at.Project(name="test", framework=F, databook=D.to_spreadsheet(), do_run=True)

    res = P.results[0]

    assert res.get_variable("c1")[0].vals[1] == 200 * res.dt  # 200 inflow, default dt=0.25
    assert res.get_variable(":j1")[0].vals[1] == 200 * res.dt
    assert res.get_variable(":j2")[0].vals[1] == 0.6 * 200 * res.dt
    assert res.get_variable(":j3")[0].vals[1] == 0.4 * 200 * res.dt
    assert res.get_variable("c2")[0].vals[1] == 0
    assert res.get_variable("c2")[0].vals[2] == 200 * res.dt

    # Timed transition into C3
    assert res.get_variable("c3")[0].vals[20] == 0
    assert res.get_variable("c3")[0].vals[21] == 200 * res.dt


def test_junction_timed_remainder_flush():
    """
    Test running the model with a timed transition with 0 duration
    """

    F = at.ProjectFramework(testdir / "framework_junction_timed_remainder_test.xlsx")
    D = at.ProjectData.new(F, [2018], pops=1, transfers=0)
    D.tdve["state_dur"].ts[0].insert(None, 0)  # Override with zero duration
    P = at.Project(name="test", framework=F, databook=D.to_spreadsheet(), do_run=False)
    res = P.run_sim()

    assert np.all(res.get_variable("c2")[0].vals == 0)  # With 0 timed duration, everyone must leave C1 immediately
    assert res.get_variable("c1")[0].vals[0] == 0
    assert res.get_variable("c1")[0].vals[1] == 50
    assert res.get_variable("c1")[0].vals[2] == 50
    assert res.get_variable("c3")[0].vals[0] == 0
    assert res.get_variable("c3")[0].vals[1] == 0
    assert res.get_variable("c3")[0].vals[2] == 50
    assert res.get_variable("c3")[0].vals[3] == 100


def test_junction_timed_remainder_single():
    """
    Test running the model with a timed transition and only one timestep
    """

    F = at.ProjectFramework(testdir / "framework_junction_timed_remainder_test.xlsx")
    D = at.ProjectData.new(F, [2018], pops=1, transfers=0)
    P = at.Project(name="test", framework=F, databook=D.to_spreadsheet(), do_run=False)
    P.settings.update_time_vector(2018, 2019, 1)
    res = P.run_sim()
    assert res.get_variable("c1")[0].vals[0] == 0
    assert res.get_variable("c1")[0].vals[1] == 200
    res.get_variable("c1")[0]._vals.shape == (5, 2)


def test_junction_timed_remainder_zero():
    """
    Test running the model with a timed transition and only initialization (no steps)
    """

    F = at.ProjectFramework(testdir / "framework_junction_timed_remainder_test.xlsx")
    D = at.ProjectData.new(F, [2018], pops=1, transfers=0)
    P = at.Project(name="test", framework=F, databook=D.to_spreadsheet(), do_run=False)
    P.settings.update_time_vector(2018, 2018, 1)
    res = P.run_sim()
    assert res.get_variable("c1")[0].vals[0] == 0
    res.get_variable("c1")[0]._vals.shape == (5, 1)


def test_junction_feed_forward():
    """
    Test running the model with a timed transition and only initialization (no steps)
    """

    F = at.ProjectFramework(testdir / "framework_junction_feed_forward_test.xlsx")
    D = at.ProjectData.new(F, [2018], pops=1, transfers=0)
    P = at.Project(name="test", framework=F, databook=D.to_spreadsheet(), do_run=False)
    res = P.run_sim()

    assert res.get_variable("c1")[0].vals[1] == 200 * res.dt  # 200 inflow, default dt=0.25
    assert res.get_variable(":j1")[0].vals[1] == 200 * res.dt
    assert res.get_variable(":j2")[0].vals[1] == 200 * res.dt
    assert res.get_variable("c2")[0].vals[1] == 0
    assert res.get_variable("c2")[0].vals[2] == 200 * res.dt


def test_junction_feed_forward_timed():
    """
    Test running the model with a timed transition and only initialization (no steps)
    """

    F = at.ProjectFramework(testdir / "framework_junction_feed_forward_timed_test.xlsx")
    D = at.ProjectData.new(F, [2018], pops=1, transfers=0)
    P = at.Project(name="test", framework=F, databook=D.to_spreadsheet(), do_run=False)
    res = P.run_sim()

    assert res.get_variable("c1")[0].vals[1] == 200 * res.dt  # 200 inflow, default dt=0.25
    assert res.get_variable(":j1")[0].vals[1] == 200 * res.dt
    assert res.get_variable(":j2")[0].vals[1] == 200 * res.dt
    assert res.get_variable("c2")[0].vals[1] == 0
    assert res.get_variable("c2")[0].vals[2] == 200 * res.dt


if __name__ == "__main__":
    test_junctions()
    test_only_junctions()
    test_junction_remainder()
    test_junction_remainder_2()
    test_junction_timed_remainder()
    test_junction_timed_remainder_flush()
    test_junction_timed_remainder_single()
    test_junction_timed_remainder_zero()
    test_junction_feed_forward()
    test_junction_feed_forward_timed()



"""
Classes for implementing Programs

This module principally defines the Program and ProgramSet classes, which are
used to define a single program/modality (e.g., FSW programs) and a
set of programs, respectively.

"""

import io
from datetime import timezone

import numpy as np
import xlsxwriter as xw
from numpy import inf, array, exp
from xlsxwriter.utility import xl_rowcol_to_cell as xlrc

import sciris as sc
from .excel import standard_formats, apply_widths, update_widths, read_tables, TimeDependentValuesEntry, validate_category
from .system import logger, FrameworkSettings as FS
from .utils import NamedItem, TimeSeries
from .version import version, gitinfo

__all__ = ["ProgramInstructions", "ProgramSet", "Program", "Covout", "InvalidProgramBook"]


class InvalidProgramBook(Exception):
    pass


class ProgramInstructions:
    """
    Store instructions for applying programs

    A :class:`ProgramSet` contains a Python representation of the program book, with
    a collection of programs, their effects, and quantities like historical spending.
    However, to run a simulation with programs, additional information is required - for
    example, which year to switch from databook parameters to program-computed parameters.
    This type of information is specific to the simulation being run, rather than being an
    intrinsic property of a set of programs. Therefore, that information is stored in
    a :class:`ProgramInstructions` instance.

    At minimum, the :class:`ProgramInstructions` must contain the year to turn on programs.
    It can also optionally contain the year to turn off programs. In addition to the start/stop
    years, the :class:`ProgramInstructions` also contains any overwrites that should be applied
    to

    - Spending (in units of people/year)
    - Capacity (in units of people/year)
    - Fraction/proportion coverage
        - For continuous programs, specified as a dimensionless fraction
        - For one-off programs, specified as a 'fraction/year' coverage

    which thus provides the underlying implementation for program-related scenarios. The
    :class:`ProgramSet` and :class:`Program` methods access and use the :class:`ProgramInstructions`
    instances. The overwrites for spending in particular are widely used - for example, budget
    optimization is a mapping from one set of :class:`ProgramInstructions` to another, and thus
    during optimization, :class:`ProgramInstructions` instances are used to test different allocations.

    Note that the program calculation proceeds by

    1. Using spending and unit cost to compute capacity
    2. Using capacity and the compartment sizes to compute fractional coverage
    3. Using fractional coverage to compute program outcomes

    Overwrites to each quantity (spending, capacity, coverage) are applied at their
    respective stages, so if the :class:`ProgramInstructions` contains more than one
    type of overwrite, they will be applied in this same order (and later stages
    will take precedence e.g. if both capacity and coverage are overwritten, it will be
    the coverage overwrite that impacts the final parameter value).

    Finally, for simplicity, if an overwrite is provided, it entirely replaces the values
    in the program book, in contrast to parameter scenarios, which contain more complex logic
    for interpolating between databook and scenario values.

    :param start_year: Year to switch to program-calculated parameters
    :param stop_year: Year to switch back to databook parameters
    :param alloc: The allocation. It can be
              - A dict keyed by program name, containing a scalar spend, or a TimeSeries of spending values. If the spend is
                scalar, it will be assigned to the start year
              - A ProgramSet instance, in which case an allocation will be assigned by interpolating the ProgramSet's
                spending onto the program start year. This is a shortcut to ensure that budget scenarios and optimizations
                where spending is specified in future years ramp correctly from the program start year (and not the last year
                that data was entered for)
    :param capacity: Overwrites to capacity. This is a dict keyed by program name, containing a scalar capacity or a TimeSeries of capacity values
                     For convenience, the capacity overwrite should be in units of 'people/year' and it will be automatically
                     converted to a per-timestep value based on the units for the program's unit cost.
    :param coverage: Overwrites to proportion coverage. This is a dict keyed by program name, containing a scalar coverage or a TimeSeries
                     of coverage values. The overwrite is specified in dimensionless units for continuous programs, and '/year' units for
                     one-off programs.

    """

    def __init__(self, start_year: float, stop_year: float = None, alloc=None, coverage: dict = None, capacity: dict = None):
        self.start_year = start_year
        self.stop_year = stop_year if stop_year else inf

        self.alloc = sc.odict()
        if isinstance(alloc, ProgramSet):
            for prog in alloc.programs.values():
                self.alloc[prog.name] = TimeSeries(t=self.start_year, vals=prog.spend_data.interpolate(self.start_year, method="previous"))
        elif alloc:
            for prog_name, spending in alloc.items():
                if isinstance(spending, TimeSeries) and spending.has_data:
                    self.alloc[prog_name] = sc.dcp(spending)
                elif spending is not None:
                    self.alloc[prog_name] = TimeSeries(t=self.start_year, vals=spending)

        self.capacity = sc.odict()  # Dict keyed by program name that stores a time series of capacities
        if capacity:
            for prog_name, vals in capacity.items():
                if isinstance(vals, TimeSeries):
                    self.capacity[prog_name] = sc.dcp(vals)
                else:
                    self.capacity[prog_name] = TimeSeries(t=self.start_year, vals=vals)

        self.coverage = sc.odict()  # Dict keyed by program name that stores a time series of coverages

        if coverage:
            for prog_name, vals in coverage.items():
                if isinstance(vals, TimeSeries):
                    self.coverage[prog_name] = sc.dcp(vals)
                else:
                    self.coverage[prog_name] = TimeSeries(t=self.start_year, vals=vals)

    def scale_alloc(self, scale_factor: float) -> None:
        """
        Scale allocation by a constant

        This method multiplies the budget by a constant to scale the total budget up or down.
        The scale factor is applied at all time points.

        :param scale_factor: Multiplicative factor for spending
        :return: A new, scaled copy of the instructions

        """
        assert scale_factor >= 0, "Cannot have a negative scale factor"
        new = sc.dcp(self)
        for ts in new.alloc.values():
            ts.vals = [x * scale_factor for x in ts.vals]
            ts.assumption = ts.assumption * scale_factor if ts.assumption is not None else None
            ts.sigma = ts.sigma * scale_factor if ts.sigma is not None else None
        return new


class ProgramSet(NamedItem):
    """
    Representation of a single program

    A ProgramSet object is the code representation of a program book. It provides an interface for reading and
    writing the program book, as well as retrieving program outcomes (due to interactions between programs, they
    must be computed using the entire collection of programs).

    :param name: Optionally specify the name of the ProgramSet
    :param tvec: Optionally specify the years for data entry

    """

    def __init__(self, name="default", framework=None, data=None, tvec=None):

        assert framework is not None, "Must specify framework and data when instantiating a ProgramSet"
        assert data is not None, "Must specify framework and data when instantiating a ProgramSet"

        NamedItem.__init__(self, name)
        self.version = version  # Track versioning information
        self.gitinfo = gitinfo

        self.tvec = tvec  # This is the data tvec that will be used when writing the progset to a spreadsheet

        # Programs and effects
        self.programs = sc.odict()  # Stores the information on the 'targeting' and 'spending data' sheet
        self.covouts = sc.odict()  # Stores the information on the 'program effects' sheet

        # Populations, parameters, and compartments - these are all the available ones printed when writing a progbook
        # They are all of the form {code_name:{'label':full_name, 'type':pop_type}}
        self.pops = sc.dcp(data.pops)  # ProjectData already stores dicts with 'label' and 'type' keys

        # Get comps from framework
        self.comps = sc.odict()
        for _, spec in framework.comps.iterrows():
            if spec["is source"] == "y" or spec["is sink"] == "y" or spec["is junction"] == "y":
                self.comps[spec.name] = {"label": spec["display name"], "type": spec["population type"], "non_targetable": True}
            else:
                self.comps[spec.name] = {"label": spec["display name"], "type": spec["population type"], "non_targetable": False}

        # Get pars from framework
        self.pars = sc.odict()
        for _, spec in framework.pars.iterrows():
            if spec["targetable"] == "y":
                self.pars[spec.name] = {"label": spec["display name"], "type": spec["population type"]}

        self._pop_types = list(framework.pop_types.keys())

        # Metadata
        self.created = sc.now(utc=True)
        self.modified = sc.now(utc=True)
        self.currency = "$"  # The symbol for currency that will be used in the progbook

        # Internal caches
        self._book = None
        self._formats = None
        self._references = None

    def __setstate__(self, d):
        from .migration import migrate

        self.__dict__ = d
        progset = migrate(self)
        self.__dict__ = progset.__dict__

    def __repr__(self):
        output = sc.prepr(self)
        output += "    Program set name: %s\n" % self.name
        output += "            Programs: %s\n" % [prog for prog in self.programs]
        output += "        Date created: %s\n" % sc.getdate(self.created.replace(tzinfo=timezone.utc).astimezone(tz=None), dateformat="%Y-%b-%d %H:%M:%S %Z")
        output += "       Date modified: %s\n" % sc.getdate(self.modified.replace(tzinfo=timezone.utc).astimezone(tz=None), dateformat="%Y-%b-%d %H:%M:%S %Z")
        output += "============================================================\n"
        return output

    def _get_code_name(self, name: str) -> str:
        """
        Return code name given code or full name

        This function converts the input name to a code name, thereby allowing operations
        like removing pops/comps/pars/progs to be performed by code name or by full name

        Code names are unique because they key the program dictionary. Full names are likely unique
        but may not be. Code names will be returned as-is so take precedence over full names. For example,
        if you have a program with a particular code name, and other programs with that same name used as
        the full name, this method will return the program with the matching code name.

        :param name: A code name or full name
        :return: A code name

        """

        if name in self.pars or name in self.comps or name in self.pops or name in self.programs:
            return name

        for code_name, spec in self.pops.items():
            if name == spec["label"]:
                return code_name

        for code_name, spec in self.comps.items():
            if name == spec["label"]:
                return code_name

        for code_name, spec in self.pars.items():
            if name == spec["label"]:
                return code_name

        for prog in self.programs.values():
            if name == prog.label:
                return prog.name

        raise Exception('Could not find full name for quantity "%s" (n.b. this is case sensitive)' % (name))

    def add_program(self, code_name: str, full_name: str) -> None:
        """
        Add a program to the ProgramSet

        :param code_name: The code name of the new program
        :param full_name: The full name of the new program

        """
        # To add a program, we just need to construct one
        prog = Program(name=code_name, label=full_name, currency=self.currency)
        if prog.name in self.programs:
            raise Exception('Program with name "%s" is already present in the ProgramSet' % (prog.name))
        self.programs[prog.name] = prog

    def remove_program(self, name: str) -> None:
        """
        Remove a program from the ProgramSet

        :param name: The code name or full name of the program to remove

        """
        # Remove the program from the progs dict
        code_name = self._get_code_name(name)

        del self.programs[code_name]
        # Remove affected covouts
        for par in self.pars:
            for pop in self.pops:
                if (par, pop) in self.covouts and code_name in self.covouts[(par, pop)].progs:
                    del self.covouts[(par, pop)].progs[code_name]

    def add_pop(self, code_name: str, full_name: str, pop_type: str = None) -> None:
        """
        Add a population to the ``ProgramSet``

        At this level, we can add any arbitrarily named population. When the progbook is loaded in,
        the populations contained in the ``ProgramSet`` will be validated against the specified ProjectData
        instance.

        :param code_name: The code name of the new population
        :param full_name: The full name of the new population

        """

        if pop_type is None:
            pop_type = self._pop_types[0]

        self.pops[code_name] = {"label": full_name, "type": pop_type}

    def remove_pop(self, name: str) -> None:
        """
        Remove a population from the ``ProgramSet``

        This method will remove the population from the ProgramSet, as well as
        from all program target pops and will also remove the covout objects
        for the affected population.

        :param name: Code name or full name of the population to remove

        """

        code_name = self._get_code_name(name)

        for prog in self.programs.values():
            if code_name in prog.target_pops:
                prog.target_pops.remove(code_name)
            if (prog.name, code_name) in self.covouts:
                self.covouts.pop((prog.name, code_name))

        del self.pops[code_name]

    def add_comp(self, code_name: str, full_name: str, pop_type: str = None) -> None:
        """
        Add a compartment

        The primary use case would be when an existing project has a change made to the framework and it's
        desired to update an already filled out program book.

        :param code_name: Code name of the compartment to add
        :param full_name: Full name of the compartment to add

        """
        if pop_type is None:
            pop_type = self._pop_types[0]

        self.comps[code_name] = {"label": full_name, "type": pop_type}

    def remove_comp(self, name: str) -> None:
        """
        Remove a compartment

        The primary use case would be when an existing project has a change made to the framework and it's
        desired to update an already filled out program book.

        Note that removing a compartment also requires removing it as a target compartment from all
        programs, which is automatically handled by this method.

        :param code_name: Code name or full name of the compartment to remove

        """

        code_name = self._get_code_name(name)

        for prog in self.programs.values():
            if code_name in prog.target_comps:
                prog.target_comps.remove(code_name)
        del self.comps[code_name]

    def add_par(self, code_name: str, full_name: str, pop_type: str = None) -> None:
        """
        Add a parameter

        The primary use case would be when an existing project has a change made to the framework and it's
        desired to update an already filled out program book.

        :param code_name: Code name of the parameter to add
        :param full_name: Full name of the parameter to add
        :param pop_type: Code name of the population type

        """

        if pop_type is None:
            pop_type = self._pop_types[0]

        # add an impact parameter
        # a new impact parameter won't have any covouts associated with it, and no programs will be bound to it
        # So all we have to do is add it to the list
        self.pars[code_name] = {"label": full_name, "type": pop_type}

    def remove_par(self, name: str) -> None:
        """
        Remove a parameter

        The primary use case would be when an existing project has a change made to the framework and it's
        desired to update an already filled out program book.

        Note that removing a parameter also requires removing all of the :class:`Covout` instances associated with it.

        :param name: Code name or full name of the parameter to remove

        """

        code_name = self._get_code_name(name)

        for pop in self.pops:
            if (code_name, pop) in self.covouts:
                del self.covouts[(code_name, pop)]
        del self.pars[code_name]

    #######################################################################################################
    # Methods for data I/O
    #######################################################################################################

    @staticmethod
    def _normalize_inputs(framework, data, project) -> tuple:
        """
        Normalize constructor inputs

        A :class:`ProjectFramework` is constructed against a particular framework
        and data. For convenience, these can be specified by passing in a Project
        containing the framework and data. This function takes in all of the inputs
        provided to the constructor (whether making a blank instance or reading a
        spreadsheet) and returns the framework and data. The order of precedence is

        - If separate framework or data is provided, it will be used
        - Otherwise, they will be drawn from the project

        So for example, a project and data could be provided, in which case the framework
        would come from the project and the data would come from the explicit argument
        (even if the project also contained data).

        :param framework: Optionally a :class:`ProjectFramework` instance
        :param data: Optionally a :class:`ProjectData` instance
        :param project: Optionally a :class:`Project` instance
        :return: Tuple containing ``(framework,data)``

        """

        if (framework is None and project is None) or (data is None and project is None):
            errormsg = "To read in a ProgramSet, please supply one of the following sets of inputs: (a) a Framework and a ProjectData, (b) a Project."
            raise Exception(errormsg)

        if framework is None:
            if project.framework is None:
                errormsg = "A Framework was not provided, and the Project has not been initialized with a Framework"
                raise Exception(errormsg)
            else:
                framework = project.framework

        if data is None:
            if project.data is None:
                errormsg = "Project data has not been loaded yet"
                raise Exception(errormsg)
            else:
                data = project.data

        return framework, data

    @classmethod
    def from_spreadsheet(cls, spreadsheet=None, framework=None, data=None, project=None, name=None, _allow_missing_data=False):
        """
        Instantiate a ProgramSet from a spreadsheet

        To load a spreadsheet, need to either pass in

            - A Project containing a framework and data
            - ProjectFramework and ProjectData instances without a project

        :param spreadsheet: A string file path, or an sc.Spreadsheet
        :param framework: A :py:class:`ProjectFramework` instance
        :param data: A :py:class:`ProjectData` instance
        :param project: A :py:class:`Project` instance
        :param name: Optionally specify the name of the ProgramSet to create
        :param _allow_missing_data: Internal only - optionally allow missing unit costs and spending (used for loading template progbook files)
        :return: A :class:`ProgramSet`

        """
        import openpyxl

        framework, data = cls._normalize_inputs(framework, data, project)

        # Populate the available pops, comps, and pars based on the framework and data provided at this step
        self = cls(name=name, framework=framework, data=data)

        # Create and load spreadsheet
        if not isinstance(spreadsheet, sc.Spreadsheet):
            spreadsheet = sc.Spreadsheet(spreadsheet)

        workbook = openpyxl.load_workbook(spreadsheet.tofile(), read_only=True, data_only=True)  # Load in read-only mode for performance, since we don't parse comments etc.
        validate_category(workbook, "atomica:progbook")

        # Load individual sheets
        try:
            self._read_targeting(workbook["Program targeting"], framework=framework)
        except Exception as e:
            message = 'Error on sheet "Program targeting"'
            raise InvalidProgramBook("%s -> %s" % (message, e)) from e

        try:
            self._read_spending(workbook["Spending data"], _allow_missing_data=_allow_missing_data)
        except Exception as e:
            message = 'Error on sheet "Spending data"'
            raise InvalidProgramBook("%s -> %s" % (message, e)) from e

        try:
            self._read_effects(workbook["Program effects"], framework=framework, data=data)
        except Exception as e:
            message = 'Error on sheet "Program effects"'
            raise InvalidProgramBook("%s -> %s" % (message, e)) from e

        return self

    def to_workbook(self) -> tuple:
        """
        Return an open workbook for the ProgramSet

        This allows the xlsxwriter workbook to be manipulated prior to closing the
        filestream e.g. to append extra sheets. This prevents issues related to cached
        data values when reloading a workbook to append or modify content

        Warning - the workbook is backed by a BytesIO instance and needs to be closed.
        See the usage of this method in the :meth`to_spreadsheet` function.

        :return: A tuple (bytes, workbook) with a BytesIO instance and a corresponding *open* xlsxwriter workbook instance

        """
        f = io.BytesIO()  # Write to this binary stream in memory
        wb = xw.Workbook(f)

        self._book = wb
        self._book.set_properties({"category": "atomica:progbook"})
        self._formats = standard_formats(self._book)
        self._references = {}  # Reset the references dict

        self._write_targeting()
        self._write_spending()
        self._write_effects()

        self._book = None
        self._formats = None
        self._references = None

        return f, wb

    def to_spreadsheet(self) -> sc.Spreadsheet:
        """
        Return content as a Sciris Spreadsheet

        :return: A :class:`sciris.Spreadsheet` instance

        """

        f, wb = self.to_workbook()
        wb.close()  # Close the workbook to flush any xlsxwriter content
        spreadsheet = sc.Spreadsheet(f)  # Wrap it in a spreadsheet instance
        return spreadsheet

    def save(self, fname):
        ss = self.to_spreadsheet()
        ss.save(fname)

    def _read_targeting(self, sheet, framework) -> None:
        # This function reads a targeting sheet and instantiates all of the programs with appropriate targets, putting them
        # into `self.programs`
        tables, start_rows = read_tables(sheet)  # NB. only the first table will be read, so there can be other tables for comments on the first page
        self.programs = sc.odict()
        sup_header = [x.value.lower().strip() if sc.isstring(x.value) else x.value for x in tables[0][0]]
        headers = [x.value.lower().strip() if sc.isstring(x.value) else x.value for x in tables[0][1]]

        # Get the indices where the pops and comps start
        pop_start_idx = sup_header.index("targeted to (populations)")
        comp_start_idx = sup_header.index("targeted to (compartments)")

        # Check the first two columns are as expected
        assert headers[0] == "abbreviation"
        assert headers[1] == "display name"

        # Now, prepare the pop and comp lookups
        pop_idx = dict()  # Map table index to pop full name
        for i in range(pop_start_idx, comp_start_idx):
            if headers[i]:
                pop_idx[i] = headers[i]

        comp_idx = dict()  # Map table index to comp full name
        for i in range(comp_start_idx, len(headers)):
            if headers[i]:
                comp_idx[i] = headers[i]

        pop_codenames = {v["label"].lower().strip(): x for x, v in self.pops.items()}
        comp_codenames = {v["label"].lower().strip(): x for x, v in self.comps.items()}

        for row in tables[0][2:]:  # For each program to instantiate
            target_pops = []
            target_comps = []

            for i in range(pop_start_idx, comp_start_idx):
                if row[i].value and sc.isstring(row[i].value) and row[i].value.lower().strip() == "y":
                    if pop_idx[i] not in pop_codenames:
                        message = "There was a mismatch between the populations in the databook and the populations in the program book"
                        message += '\nThe program book contains population "%s", while the databook contains: %s' % (pop_idx[i], [str(x) for x in pop_codenames.keys()])
                        raise Exception(message)

                    target_pops.append(pop_codenames[pop_idx[i]])  # Append the pop's codename

            for i in range(comp_start_idx, len(headers)):
                if row[i].value and sc.isstring(row[i].value) and row[i].value.lower().strip() == "y":
                    if comp_idx[i] not in comp_codenames:
                        if comp_idx[i] in set(framework.comps["display name"].str.lower().str.strip()):
                            # It's a valid compartment of some sort (not actually targetable - for use with coverage scenarios only)
                            logger.warning(f'Compartment "{comp_idx[i]}" is a non-standard compartment that cannot be targeted. The ProgramSet can be used with coverage scenarios only')
                            spec = framework.comps.loc[framework.comps["display name"].str.lower().str.strip() == comp_idx[i]]
                            self.add_comp(code_name=spec.index[0], full_name=spec["display name"][0], pop_type=spec["population type"][0])
                            target_comps.append(spec.index[0])  # Append the pop's codename
                        else:
                            message = "There was a mismatch between the compartments in the databook and the compartments in the Framework file"
                            message += '\nThe program book contains compartment "%s", while the Framework contains: %s' % (comp_idx[i], [str(x) for x in comp_codenames.keys()])
                            raise Exception(message)
                    else:
                        target_comps.append(comp_codenames[comp_idx[i]])  # Append the pop's codename

            short_name = row[0].value.strip()
            if short_name.lower() == "all":
                raise Exception('A program was named "all", which is a reserved keyword and cannot be used as a program name')
            long_name = row[1].value.strip()

            self.programs[short_name] = Program(name=short_name, label=long_name, target_pops=target_pops, target_comps=target_comps)

    def _write_targeting(self):
        sheet = self._book.add_worksheet("Program targeting")
        widths = dict()

        # Work out the column offset associated with each population label and comp label
        pop_block_offset = 2  # This is the co
        sheet.write(0, pop_block_offset, "Targeted to (populations)", self._formats["rc_title"]["left"]["F"])
        comp_block_offset = 2 + len(self.pops) + 1
        sheet.write(0, comp_block_offset, "Targeted to (compartments)", self._formats["rc_title"]["left"]["F"])

        comps_in_use = set()
        for prog in self.programs.values():
            comps_in_use.update(prog.target_comps)
        comps_to_write = {k: v for k, v in self.comps.items() if k in comps_in_use or not v["non_targetable"]}

        pop_col = {n: i + pop_block_offset for i, n in enumerate(self.pops.keys())}
        comp_col = {n: i + comp_block_offset for i, n in enumerate(comps_to_write.keys())}

        # Write the header row
        sheet.write(1, 0, "Abbreviation", self._formats["center_bold"])
        update_widths(widths, 0, "Abbreviation")
        sheet.write(1, 1, "Display name", self._formats["center_bold"])
        update_widths(widths, 1, "Abbreviation")
        for pop, spec in self.pops.items():
            col = pop_col[pop]
            sheet.write(1, col, spec["label"], self._formats["rc_title"]["left"]["T"])
            self._references[spec["label"]] = "='%s'!%s" % (sheet.name, xlrc(1, col, True, True))
            widths[col] = 12  # Wrap population names

        for comp, spec in comps_to_write.items():
            if spec["non_targetable"] and comp not in comps_in_use:
                continue
            col = comp_col[comp]
            sheet.write(1, col, spec["label"], self._formats["rc_title"]["left"]["T"])
            self._references[spec["label"]] = "='%s'!%s" % (sheet.name, xlrc(1, col, True, True))
            widths[col] = 12  # Wrap compartment names

        row = 2
        self._references["reach_pop"] = dict()  # This is storing cells e.g. self._references['reach_pop'][('BCG','0-4')]='$A$4' so that conditional formatting can be done
        for prog in self.programs.values():
            sheet.write(row, 0, prog.name)
            self._references[prog.name] = "='%s'!%s" % (sheet.name, xlrc(row, 0, True, True))
            update_widths(widths, 0, prog.name)
            sheet.write(row, 1, prog.label)
            self._references[prog.label] = "='%s'!%s" % (sheet.name, xlrc(row, 1, True, True))
            update_widths(widths, 1, prog.label)

            for pop in self.pops.keys():
                col = pop_col[pop]
                if pop in prog.target_pops:
                    sheet.write(row, col, "Y", self._formats["center"])
                else:
                    sheet.write(row, col, "N", self._formats["center"])
                self._references["reach_pop"][(prog.name, pop)] = "'%s'!%s" % (sheet.name, xlrc(row, col, True, True))
                sheet.data_validation(xlrc(row, col), {"validate": "list", "source": ["Y", "N"]})
                sheet.conditional_format(xlrc(row, col), {"type": "cell", "criteria": "equal to", "value": '"Y"', "format": self._formats["unlocked_boolean_true"]})
                # sheet.conditional_format(xlrc(row, col), {'type': 'cell', 'criteria': 'equal to', 'value': '"N"', 'format': self._formats['unlocked_boolean_false']})

            for comp in comps_to_write.keys():
                col = comp_col[comp]
                if comp in prog.target_comps:
                    sheet.write(row, col, "Y", self._formats["center"])
                else:
                    sheet.write(row, col, "N", self._formats["center"])
                sheet.data_validation(xlrc(row, col), {"validate": "list", "source": ["Y", "N"]})
                sheet.conditional_format(xlrc(row, col), {"type": "cell", "criteria": "equal to", "value": '"Y"', "format": self._formats["unlocked_boolean_true"]})
                # sheet.conditional_format(xlrc(row, col), {'type': 'cell', 'criteria': 'equal to', 'value': '"N"', 'format': self._formats['unlocked_boolean_false']})

            row += 1

        apply_widths(sheet, widths)

    def _read_spending(self, sheet, _allow_missing_data=False):
        """
        Internal method to read the spending data

        :param sheet:
        :param _allow_missing_data: If False, an error will be raised if unit costs or total spending is missing.
                                   However, the FE requires loading in a progset with missing spending data. In that case,
                                   having these values missing is be allowed
        :return:
        """
        # Read the spending table and populate the program data
        tables, start_rows = read_tables(sheet)
        times = set()
        for table, start_row in zip(tables, start_rows):
            try:
                tdve = TimeDependentValuesEntry.from_rows(table)
            except Exception as e:
                message = 'Error on sheet "%s" while trying to read a TDVE table starting on row %d -> ' % (sheet.title, start_row)
                raise Exception("%s -> %s" % (message, e)) from e

            prog = self.programs[tdve.name]

            def set_ts(prog, field_name, ts):
                # Set Program timeseries based on TDVE, handling the case where the TDVE has no units
                if ts.units is None:
                    ts.units = getattr(prog, field_name).units
                setattr(prog, field_name, ts)

            if "Total spend" in tdve.ts:  # This is a compatibility statement around 19/10/18, progbooks generated during the WB workshop would have 'Total spend'
                set_ts(prog, "spend_data", tdve.ts["Total spend"])
            else:
                set_ts(prog, "spend_data", tdve.ts["Annual spend"])

            if "Capacity" in tdve.ts:  # Old progbooks have 'capacity' instead of 'capacity constraint'
                set_ts(prog, "capacity_constraint", tdve.ts["Capacity"])
            else:
                set_ts(prog, "capacity_constraint", tdve.ts["Capacity constraint"])

            set_ts(prog, "unit_cost", tdve.ts["Unit cost"])
            set_ts(prog, "coverage", tdve.ts["Coverage"])
            set_ts(prog, "saturation", tdve.ts["Saturation"])

            if not _allow_missing_data:
                assert prog.unit_cost.has_data, 'Unit cost data for %s not was not entered (in table on sheet "%s" starting on row %d' % (prog.name, sheet.title, start_row)
                assert prog.spend_data.has_data, 'Spending data for %s not was not entered (in table on sheet "%s" starting on row %d' % (prog.name, sheet.title, start_row)

            if "/year" in prog.unit_cost.units and "/year" in prog.coverage.units:
                logger.warning("Program %s: Typically if the unit cost is `/year` then the coverage would not be `/year`", prog.label)
            times.update(set(tdve.tvec))

        # Work out the currency
        units = set([x.spend_data.units.split("/")[0].strip() for x in self.programs.values()])
        units.update([x.unit_cost.units.split("/")[0].strip() for x in self.programs.values()])

        if len(units) == 1:
            self.currency = list(units)[0]
        else:
            raise Exception("The progbook contains multiple currencies: (%s). All spending must be specified in the same currency" % units)

        self.tvec = array(sorted(list(times)))  # NB. This means that  the ProgramSet's tvec (used when writing new programs) is based on the last Program to be read in

    def _write_spending(self):
        sheet = self._book.add_worksheet("Spending data")
        widths = dict()
        next_row = 0

        for prog in self.programs.values():
            # Make a TDVE table for
            tdve = TimeDependentValuesEntry(prog.name, self.tvec)
            prog = self.programs[tdve.name]
            tdve.ts["Annual spend"] = prog.spend_data
            tdve.ts["Unit cost"] = prog.unit_cost
            tdve.ts["Capacity constraint"] = prog.capacity_constraint
            tdve.ts["Saturation"] = prog.saturation
            tdve.ts["Coverage"] = prog.coverage

            tdve.assumption_heading = "Assumption"
            tdve.write_assumption = True
            tdve.write_units = True
            tdve.write_uncertainty = True

            tdve.allowed_units = {"Unit cost": [self.currency + "/person (one-off)", self.currency + "/person/year"], "Capacity": ["people/year", "people"]}

            # NOTE - If the ts contains time values that aren't in the ProgramSet's tvec, then an error will be thrown
            # However, if the ProgramSet's tvec contains values that the ts does not, then that's fine, there
            # will just be an empty cell in the spreadsheet
            next_row = tdve.write(sheet, next_row, self._formats, self._references, widths)

        apply_widths(sheet, widths)

    def _read_effects(self, sheet, framework, data):
        # Read the program effects sheet. Here we instantiate a costcov object for every non-empty row

        tables, start_rows = read_tables(sheet)
        pop_codenames = {v["label"].lower().strip(): x for x, v in self.pops.items()}
        par_codenames = {v["label"].lower().strip(): x for x, v in self.pars.items()}
        transfer_names = set()
        for transfer in data.transfers:
            for pops in transfer.ts.keys():
                transfer_names.add(("%s_%s_to_%s" % (transfer.code_name, pops[0], pops[1])).lower())

        self.covouts = sc.odict()

        for table in tables:
            par_label = table[0][0].value.strip()
            if par_label.lower() in par_codenames:
                par_name = par_codenames[par_label.lower()]  # Code name of the parameter we are working with
            elif par_label.lower() in transfer_names:
                par_name = table[0][0].value.strip()  # Preserve case
            else:
                raise Exception('Program name "%s" was not found in the framework parameters or in the databook transfers' % (table[0][0].value.strip()))

            # Read the headers - on this sheet, string type only
            idx_to_header = {}
            for i, cell in enumerate(table[0]):
                v = cell.value
                if sc.isstring(v):
                    if v.startswith("#ignore"):
                        break
                    else:
                        idx_to_header[i] = v.strip()
                elif cell.value is None:
                    continue
                else:
                    raise Exception(f"Unknown data type in cell {cell.coordinate}")

            for row in table[1:]:
                # For each covout row, we will initialize
                pop_label = row[0].value.strip()
                if pop_label.lower() not in pop_codenames:
                    raise Exception(f'Population "{pop_label}" was not found in the databook')
                pop_name = pop_codenames[pop_label.lower()]

                # Code name of the population we are working on
                if self.pars[par_name]["type"] != self.pops[pop_name]["type"]:
                    raise Exception(f'On the Effects sheet, Parameter "{par_label}" belongs to population type "{self.pars[par_name]["type"]}" but Population "{pop_label}" (Cell {row[0].coordinate}) has population type "{self.pops[pop_name]["type"]}"')

                progs = sc.odict()

                baseline = None
                cov_interaction = None
                imp_interaction = None
                uncertainty = None

                for i, x in enumerate(row[1:]):
                    i = i + 1  # Offset of 1 because the loop is over row[1:] not row[0:]

                    try:
                        if idx_to_header.get(i, None) is None:  # If the header row had a blank cell, ignore everything in that column - we don't know what it is otherwise
                            continue
                        elif idx_to_header[i].lower() == "baseline value":
                            if x.value is not None:  # test `is not None` because it might be entered as 0
                                baseline = float(x.value)
                        elif idx_to_header[i].lower() == "coverage interaction":
                            if x.value:
                                cov_interaction = x.value.strip().lower()  # additive, nested, etc.
                        elif idx_to_header[i].lower() == "impact interaction":
                            if x.value:
                                imp_interaction = x.value.strip()  # additive, nested, etc.
                        elif idx_to_header[i].lower() == "uncertainty":
                            if x.value is not None:  # test `is not None` because it might be entered as 0
                                uncertainty = float(x.value)
                        elif x.value is not None:  # If the header isn't empty, then it should be one of the program names
                            if idx_to_header[i] not in self.programs:
                                raise Exception('The heading "%s" was not recognized as a program name or a special token - spelling error?' % (idx_to_header[i]))
                            progs[idx_to_header[i]] = float(x.value)
                    except Exception as e:
                        message = f"Error in cell {x.coordinate}"
                        raise Exception("%s -> %s" % (message, e)) from e

                if baseline is None and progs:
                    raise Exception(f'On the "Effects" sheet for Parameter "{table[0][0].value.strip()}" in population "{row[0].value.strip()}", program outcomes are defined but the baseline value is missing')
                if baseline is not None or progs:  # Only instantiate covout objects if they have programs associated with them
                    self.covouts[(par_name, pop_name)] = Covout(par=par_name, pop=pop_name, cov_interaction=cov_interaction, imp_interaction=imp_interaction, uncertainty=uncertainty, baseline=baseline, progs=progs)

    def _write_effects(self):
        sheet = self._book.add_worksheet("Program effects")
        widths = dict()

        current_row = 0

        for par_name, par_spec in self.pars.items():
            sheet.write(current_row, 0, par_spec["label"], self._formats["rc_title"]["left"]["F"])
            update_widths(widths, 0, par_spec["label"])

            for i, s in enumerate(["Baseline value", "Coverage interaction", "Impact interaction", "Uncertainty"]):
                sheet.write(current_row, 1 + i, s, self._formats["rc_title"]["left"]["T"])
                widths[1 + i] = 12  # Fixed width, wrapping
            # sheet.write_comment(xlrc(current_row,1), 'In this column, enter the baseline value for "%s" if none of the programs reach this parameter (e.g., if the coverage is 0)' % (par_label))

            applicable_progs = self.programs.values()  # All programs - could filter this later on
            prog_col = {p.name: i + 6 for i, p in enumerate(applicable_progs)}  # add any extra padding columns to the indices here too

            for prog in applicable_progs:
                sheet.write_formula(current_row, prog_col[prog.name], self._references[prog.name], self._formats["center_bold"], value=prog.name)
                update_widths(widths, prog_col[prog.name], prog.name)
            current_row += 1

            applicable_covouts = {x.pop: x for x in self.covouts.values() if x.par == par_name}
            applicable_pops = [x for x, v in self.pops.items() if v["type"] == par_spec["type"]]  # All populations with matching type

            for pop_name in applicable_pops:

                if pop_name not in applicable_covouts:  # There is currently no covout
                    covout = None
                else:
                    covout = applicable_covouts[pop_name]

                sheet.write_formula(current_row, 0, self._references[self.pops[pop_name]["label"]], value=self.pops[pop_name]["label"])
                update_widths(widths, 0, self.pops[pop_name])

                if covout and covout.baseline is not None:
                    sheet.write(current_row, 1, covout.baseline, self._formats["not_required"])
                else:
                    sheet.write(current_row, 1, None, self._formats["unlocked"])

                if covout and covout.cov_interaction is not None:
                    sheet.write(current_row, 2, covout.cov_interaction.title(), self._formats["not_required"])
                else:
                    sheet.write(current_row, 2, "Additive", self._formats["unlocked"])
                sheet.data_validation(xlrc(current_row, 2), {"validate": "list", "source": ["Random", "Additive", "Nested"]})

                if covout and covout.imp_interaction is not None:
                    sheet.write(current_row, 3, covout.imp_interaction, self._formats["not_required"])
                else:
                    sheet.write(current_row, 3, None, self._formats["unlocked"])

                if covout and covout.sigma is not None:
                    sheet.write(current_row, 4, covout.sigma, self._formats["not_required"])
                else:
                    sheet.write(current_row, 4, None, self._formats["unlocked"])

                for prog in applicable_progs:
                    if covout and prog.name in covout.progs:
                        sheet.write(current_row, prog_col[prog.name], covout.progs[prog.name], self._formats["not_required"])
                    else:
                        sheet.write(current_row, prog_col[prog.name], None, self._formats["unlocked"])

                    fcn_pop_not_reached = '%s<>"Y"' % (self._references["reach_pop"][(prog.name, pop_name)])  # Excel formula returns FALSE if pop was 'N' (or blank)
                    sheet.conditional_format(xlrc(current_row, prog_col[prog.name]), {"type": "formula", "criteria": "=AND(%s,NOT(ISBLANK(%s)))" % (fcn_pop_not_reached, xlrc(current_row, prog_col[prog.name])), "format": self._formats["ignored_warning"]})
                    sheet.conditional_format(xlrc(current_row, prog_col[prog.name]), {"type": "formula", "criteria": "=" + fcn_pop_not_reached, "format": self._formats["ignored_not_required"]})

                # Conditional formatting for the impact interaction - hatched out if no single-program outcomes
                fcn_empty_outcomes = 'COUNTIF(%s:%s,"<>" & "")<2' % (xlrc(current_row, 5), xlrc(current_row, 5 + len(applicable_progs)))
                sheet.conditional_format(xlrc(current_row, 3), {"type": "formula", "criteria": "=" + fcn_empty_outcomes, "format": self._formats["ignored"]})
                sheet.conditional_format(xlrc(current_row, 3), {"type": "formula", "criteria": "=AND(%s,NOT(ISBLANK(%s)))" % (fcn_empty_outcomes, xlrc(current_row, 3)), "format": self._formats["ignored_warning"]})

                current_row += 1

            current_row += 1

        apply_widths(sheet, widths)

    @staticmethod
    def new(name=None, tvec=None, progs=None, project=None, framework=None, data=None):
        """
        Generate a new progset with blank data

        :param name: the name for the progset
        :param tvec: an np.array() with the time values to write
        :param progs: This can be
              - A number of progs
              - An odict of {code_name:display name} programs
        :param project: specify a project to use the project's framework and data to initialize the comps, pars, and pops
        :param framework: specify a framework to use the framework's comps and pars
        :param data: specify a data to use the data's pops
        :return: A new :class:`ProgramSet` instances

        """

        assert tvec is not None, "You must specify the time points where data will be entered"
        # Prepare programs
        if sc.isnumber(progs):
            nprogs = progs
            progs = sc.odict()
            for p in range(nprogs):
                progs["Prog %i" % (p + 1)] = "Program %i" % (p + 1)
        elif isinstance(progs, dict):  # will also match odict
            pass
        elif progs is None:
            errormsg = "When creating a ProgramSet, the programs cannot be None - it can either be a number of programs, or a dict with code names and full names"
            raise Exception(errormsg)
        else:
            errormsg = 'Please just supply a number of programs, not "%s"' % (type(progs))
            raise Exception(errormsg)

        framework, data = ProgramSet._normalize_inputs(framework, data, project)  # This step will fail if the framework and data cannot be resolved

        newps = ProgramSet(name=name, tvec=tvec, framework=framework, data=data)
        if not newps.pars:
            logger.warning('No parameters were marked as "Targetable" in the Framework, so no program effects were added')

        for k, v in progs.items():
            newps.add_program(code_name=k, full_name=v)

        return newps

    def validate(self) -> None:
        """
        Perform basic validation checks

        :raises: Exception if anything is invalid

        """

        for prog in self.programs.values():
            if not prog.target_comps and self.comps:
                # If there are no compartments, then it's fine not to target any compartments - it should be obvious that only coverage scenarios are possible
                # If there are compartments, then the same is true, but it's also possible (or even probable) that the user accidentally didn't target any compartments
                # Therefore, in this case, raise an error - if a user wants to just do coverage scenarios, then they can still target the compartments anyway
                raise Exception('Program "%s" does not target any compartments' % (prog.name))
            if not prog.target_pops:
                # If the user is using parameters only, they will still have to define a population. And that population must be targeted in order
                # to provide any program outcome values. Thus, the program should generally target the population even if there are no compartments,
                # so we raise an error if no populations are targeted
                raise Exception('Program "%s" does not target any populations' % (prog.name))

    #######################################################################################################
    # Methods for getting core response summaries: budget, allocations, coverages, outcomes, etc
    #######################################################################################################

    def get_alloc(self, tvec, instructions=None) -> dict:
        """
        Return the spending allocation for each program

        This method fuses the spending data entered in the program book with
        any overwrites that are present in the instructions. The spending values
        returned by this method thus reflect any budget scenarios that may be
        present.

        :param tvec: array of times (in years) - this is required to interpolate time-varying spending values
        :param instructions: optionally specify instructions, which can supply a spending overwrite
        :return: Dict like ``{prog_name: np.array()}`` with spending on each program (in units of '$/year' or currency equivalent)

        """

        tvec = sc.promotetoarray(tvec)

        alloc = sc.odict()
        for prog in self.programs.values():
            if instructions is None or prog.name not in instructions.alloc:
                alloc[prog.name] = prog.get_spend(tvec)
            else:
                alloc[prog.name] = instructions.alloc[prog.name].interpolate(tvec, method="previous")

        if instructions:
            for prog_name in set(instructions.alloc.keys()) - set(self.programs.keys()):
                logger.warning('The instructions contain an overwrite for a program called "%s" but as this is not in the ProgramSet, it will have no effect', prog_name)

        return alloc

    def get_capacities(self, tvec, dt, instructions=None) -> dict:
        """
        Return timestep capacity for all programs

        For convenience, this method automatically calls :meth:`ProgramSet.get_alloc()` to
        retrieve the spending values that are used to compute capacity. Thus, this method
        fuses the program capacity computed with the inclusion of any budget scenarios, with
        any capacity overwrites that are present in the instructions.

        :param tvec: array of times (in years) - this is required to interpolate time-varying unit costs and capacity_constraint constraints
        :param dt: scalar timestep size - this is required to adjust spending on incidence-type programs
        :param instructions: optionally specify instructions, which can supply a spending overwrite
        :return: Dict like ``{prog_name: np.array()}`` with program capacity at each timestep (in units of people)

        """

        # Validate inputs
        tvec = sc.promotetoarray(tvec)
        alloc = self.get_alloc(tvec, instructions)

        # Get number covered for each program
        capacities = sc.odict()  # Initialise outputs
        for prog in self.programs.values():
            if instructions is None or prog.name not in instructions.capacity:
                if prog.name in alloc:
                    spending = alloc[prog.name]
                else:
                    spending = None
                # Note that prog.get_capacity() returns capacity in units of people
                capacities[prog.name] = prog.get_capacity(tvec=tvec, dt=dt, spending=spending)
            else:
                capacities[prog.name] = instructions.capacity[prog.name].interpolate(tvec, method="previous")
                # Capacity overwrites are input in units of people/year so convert to units of people here
                if prog.is_one_off:
                    capacities[prog.name] *= dt

        return capacities

    def get_prop_coverage(self, tvec, dt, capacities: dict, num_eligible: dict, instructions=None) -> dict:
        """
        Return dimensionless (timestep) fractional coverage

        Note that this function is primarily for internal usage (i.e. during
        model integration or reconciliation). Since the proportion covered depends
        on the number of people eligible for the program (the coverage denominator),
        retrieving fractional coverage after running the model is best accomplished
        via ``Result.get_coverage('fraction')`` whereas this method is called automatically
        during integration.

        Evaluating the proportion coverage for a ProgramSet is not a straight division because

        - instructions can override the coverage (for coverage scenarios)
        - Programs can contain saturation constraints

        Note that while converage overwrites might be specified in '/year' units for one-off programs,
        *this function always returns dimensionless coverage* (i.e. not coverage/year). This coverage is also capped at 1.

        :param tvec: scalar year, or array of years - this is required to interpolate time-varying saturation values
        :param capacities: dict of program coverages, should match the available programs (typically the output of ``ProgramSet.get_capacities()``)
                           Note that since the capacity and eligible compartment sizes are being compared here,
                           the capacity needs to be in units of 'people' (not 'people/year') at this point
        :param num_eligible: dict of number of people covered by each program, computed externally and with one entry for each program
        :param instructions: optionally specify instructions, which can supply a coverage overwrite
        :return: Dict like ``{prog_name: np.array()}`` with fractional coverage values (dimensionless)

        """

        prop_coverage = sc.odict()  # Initialise outputs

        for prog in self.programs.values():
            if instructions is None or prog.name not in instructions.coverage:
                # The capacities have already been converted to timestep values, so no further transformation is necessary
                prop_coverage[prog.name] = prog.get_prop_covered(tvec, capacities[prog.name], num_eligible[prog.name])
            else:
                prop_coverage[prog.name] = instructions.coverage[prog.name].interpolate(tvec, method="previous")
                if prog.is_one_off:
                    # Coverage overwrites for one off programs are specified in /year units, therefore they get adjusted by dt here
                    prop_coverage[prog.name] *= dt

            prop_coverage[prog.name] = np.minimum(prop_coverage[prog.name], 1.0)

        return prop_coverage

    def get_outcomes(self, prop_coverage: dict) -> dict:
        """
        Get program outcomes given fractional coverage

        Since the modality interactions in Covout.get_outcome() assume that the coverage is scalar, this function
        will also only work for scalar coverage. Therefore, the prop coverage would normally come from
        ProgramSet.get_prop_coverage(tvec,...) where tvec was only one year.

        Note that this function is mainly aimed at internal usage. Typically, the program-provided
        parameter values would be best accessed by examining the appropriate output in the ``Result``.
        For example, if the programs system overwrites the screening rate ``screen`` then it would
        normally be easiest to run a simulation and then use ``Result.get_variable(popname,'screen')``

        For computational efficiency, this method returns a flat dictionary keyed by
        parameter-population pairs that then gets inserted into the appropriate integration objects
        by :meth:`Model.update_pars`.

        :param prop_coverage: dict with coverage values ``{prog_name:val}``
        :return: dict ``{(par,pop):val}`` containing parameter value overwrites

        """

        return {(covout.par, covout.pop): covout.get_outcome(prop_coverage) for covout in self.covouts.values()}

    def sample(self, constant: bool = True):
        """
        Perturb programs based on uncertainties

        The covout objects contained within the ProgramSet cache the program outcomes. At construction,
        the cache corresponds to the values entered in the databook. Calling this function will perturb
        the caches based on the sigma values. A simulation subsequently run using the ProgramSet will
        use the perturbed outcomes.

        :param constant: If True, time series will be perturbed by a single constant offset. If False,
                         an different perturbation will be applied to each time specific value independently.
        :return: A new ``ProgramSet`` with values perturbed by sampling

        """

        new = sc.dcp(self)
        for prog in new.programs.values():
            prog.sample(constant)
        for covout in new.covouts.values():
            covout.sample()
        return new


class Program(NamedItem):
    """
    Representation of a single program

    A :class:`Program` object is instantiated for every program listed on the 'Program Targeting'
    sheet in the program book. The :class:`Program` object contains

    - Collections of targeted populations and compartments from the targeting sheet in the program book
    - The time-dependent program properties on the spending data sheet (such as total spend and unit cost)

    :param name: Short name of the program
    :param label: Full name of the program
    :param target_pops: List of population code names for pops targeted by the program
    :param target_comps: List of compartment code names for compartments targeted by the program
    :param currency: The currency to use (for display purposes only) - normally this would be set to ``ProgramSet.currency`` by ``ProgramSet.add_program()``

    :param name: Short name of the program
    :param label: Full name of the program
    :param target_pops: List of population code names for pops targeted by the program
    :param target_comps: List of compartment code names for compartments targeted by the program
    :param currency: The currency to use (for display purposes only) - normally this would be set to ``ProgramSet.currency`` by ``ProgramSet.add_program()``

    """

    def __init__(self, name, label=None, target_pops=None, target_comps=None, currency="$"):
        NamedItem.__init__(self, name)
        self.name = name  #: Short name of program
        self.label = name if label is None else label  #: Full name of the program
        self.target_pops = [] if target_pops is None else target_pops  #: List of populations targeted by the program
        self.target_comps = [] if target_comps is None else target_comps  #: Compartments targeted by the program - used for calculating coverage denominators
        self.baseline_spend = TimeSeries(assumption=0.0, units=currency + "/year")  #: A TimeSeries with any baseline spending data - currently not exposed in progbook
        self.spend_data = TimeSeries(units=currency + "/year")  #: TimeSeries with spending data for the program
        self.unit_cost = TimeSeries(units=currency + "/person (one-off)")  #: TimeSeries with unit cost of the program
        self.capacity_constraint = TimeSeries(units="people/year")  #: TimeSeries with capacity constraint for the program
        self.saturation = TimeSeries(units=FS.DEFAULT_SYMBOL_INAPPLICABLE)  #: TimeSeries with saturation constraint that is applied to fractional coverage
        self.coverage = TimeSeries(units="people/year")  #: TimeSeries with capacity of program - optional - if not supplied, cost function is assumed to be linear

    @property
    def is_one_off(self) -> bool:
        """
        Flag for one-off programs

        A one-off program is a program where the cost is incurred once, per person impacted. For example, a treatment
        program where after treatment the person is no longer eligible for treatment. In contrast, a non-one-off program
        (a continuous program) is one where a person reached by the program remains eligible - for example, ART. In addition,
        one-off programs are typically linked to transition parameters, while continuous programs are typically linked to
        non-transition parameters.

        Whether a program is one-off or not depends on whether the unit cost is specified as
        - Cost per person (one-off)
        - Cost per person per year (continuous)

        :return: True if program is a one-off program

        """
        return "/year" not in self.unit_cost.units

    def sample(self, constant: bool) -> None:
        """
        Perturb program values based on uncertainties

        Calling this function will perturb the original values based on their uncertainties. The
        values will change in-place. Normally, this method would be called by
        :meth:`ProgramSet.sample()` which will copy the ``Program`` instance first.

        :param constant: If True, time series will be perturbed by a single constant offset. If False,
                         an different perturbation will be applied to each time specific value independently.
        """

        self.spend_data = self.spend_data.sample(constant)
        self.unit_cost = self.unit_cost.sample(constant)
        self.capacity_constraint = self.capacity_constraint.sample(constant)
        self.saturation = self.saturation.sample(constant)
        self.coverage = self.coverage.sample(constant)

    def __repr__(self):
        output = sc.prepr(self)
        output += "          Program name: %s\n" % self.name
        output += "         Program label: %s\n" % self.label
        output += "  Targeted populations: %s\n" % self.target_pops
        output += " Targeted compartments: %s\n" % self.target_comps
        output += "\n"
        return output

    def get_spend(self, year=None, total: bool = False) -> np.array:
        """
        Retrieve program spending

        :param year: Scalar, list, or array of years to retrieve spending in
        :param total: If True, the baseline spend will be added to the spending data
        :return: Array of spending values

        """

        if total:
            return self.spend_data.interpolate(year, method="previous") + self.baseline_spend.interpolate(year, method="previous")
        else:
            return self.spend_data.interpolate(year, method="previous")

    def get_capacity(self, tvec, spending, dt):
        """
        Return timestep capacity

        This method returns the number of people covered at each timestep. For one-off
        programs, this means the annual capacity is multiplied by the timestep size.
        Whether timestep scaling takes place or not is determined based on the units
        of the unit cost ($/person or $/person/year where the former is used for one-off programs).

        The method takes in the spending value to support overwrites in spending value by program
        instructions (in which case, spending should be drawn from the instructions rather than
        the program's spending data). This is handled in :meth:`ProgramSet.get_capacities`

        This method returns the program's timestep capacity - that is, the capacity of the program per
        timestep, at specified points in time. The spending and capacity constraint are automatically
        adjusted depending on the units of the unit cost and capacity constraint such that the
        calculation returns capacity in units of 'people' (*not* 'people/year')

        :param tvec: A scalar, list, or array of times
        :param spending: A vector of spending values (in units of '$/year'), the same size as ``tvec``
        :param dt: The time step size (required because the number covered at each time step potentially depends on the spending per-timestep)
        :return: Array the same size as ``tvec``, with capacity in units of 'people'

        """

        # Validate inputs
        spending = sc.promotetoarray(spending)

        unit_cost = self.unit_cost.interpolate(tvec, method="previous")
        if self.is_one_off:
            # The spending is $/year, and the /year gets eliminated if the unit cost is also per year. For one-off programs, the unit cost is not
            # /year, therefore we need to multiply the spending by the timestep to the capacity as people rather than people/year
            spending *= dt

        capacity = spending / unit_cost

        if self.capacity_constraint.has_data:
            capacity_constraint = self.capacity_constraint.interpolate(tvec, method="previous")
            if "/year" in self.capacity_constraint.units:
                # The capacity_constraint constraint is applied to a number of people. If it is /year, then it must be multiplied by the timestep first
                capacity_constraint *= dt
            capacity = np.minimum(capacity_constraint, capacity)

        return capacity

    def get_prop_covered(self, tvec, capacity, eligible):
        """
        Return proportion of people covered

        The time vector ``tvec`` is required to interpolate the saturation values. The
        ``capacity`` and ``eligible`` variables are assumed to correspond to the same time points
        and thus the array sizes should match the size of the time array.

        :param tvec:  An array of times
        :param capacity: An array of number of people covered (e.g. the output of ``Program.get_capacity()``)
                         This should be in units of 'people', rather than 'people/year' although it may be a
                         timestep-sensitive value if the capacity follows from a timestep-adjusted spend.
        :param eligible: The number of people eligible for the program (computed from a model object or a Result)
        :return: The fractional coverage (used to compute outcomes)

        """

        tvec = sc.promotetoarray(tvec)
        eligible = sc.promotetoarray(eligible)
        capacity = sc.promotetoarray(capacity)

        if self.saturation.has_data:
            # If the coverage denominator (eligible) is 0, then we need to use the saturation value
            prop_covered = np.divide(capacity, eligible, out=np.full(capacity.shape, np.inf), where=eligible != 0)
            saturation = self.saturation.interpolate(tvec, method="previous")
            prop_covered = 2 * saturation / (1 + exp(-2 * prop_covered / saturation)) - saturation
            prop_covered = np.minimum(prop_covered, 1.0)  # Ensure that coverage doesn't go above 1 (if saturation is < 1)
        else:
            # The division below means that 0/0 is treated as returning 1
            prop_covered = np.divide(capacity, eligible, out=np.ones_like(capacity), where=eligible > capacity)

        return prop_covered


class Covout:
    """
    Store and compute program outcomes

    The :class:`Covout` object is responsible for storing the


    :param par: string with the code name of the parameter being overwritten
    :param pop: string with the code name of the population being overwritten
    :param progs: a dict containing ``{prog_name:outcome}`` with the single program outcomes
    :param cov_interaction: one of 'additive', 'random', 'nested'
    :param imp_interaction: a parsable string like ``'Prog1+Prog2=10,Prog2+Prog3=20'`` with the interaction outcomes
    :param uncertainty: a scalar standard deviation for the outcomes
    :param baseline: the zero coverage baseline value

    """

    def __init__(self, par: str, pop: str, progs: dict, cov_interaction: str = None, imp_interaction: str = None, uncertainty=0.0, baseline=0.0):

        if cov_interaction is None:
            cov_interaction = "additive"
        else:
            assert cov_interaction in ["additive", "random", "nested"], 'Coverage interaction must be set to "additive", "random", or "nested"'

        self.par = par
        self.pop = pop
        self.cov_interaction = cov_interaction
        self.imp_interaction = imp_interaction
        self.progs = sc.dcp(progs)
        self.sigma = uncertainty
        self.baseline = baseline

        # Parse the interactions into a numeric representation
        self._interactions = dict()
        if self.imp_interaction and not self.imp_interaction.lower() in ["best", "synergistic"]:
            for interaction in self.imp_interaction.split(","):
                combo, val = interaction.split("=")
                combo = frozenset([x.strip() for x in combo.split("+")])
                for x in combo:
                    assert x in self.progs, 'The impact interaction refers to a program "%s" which does not appear in the available programs' % (x)
                self._interactions[combo] = float(val) - self.baseline

        self.update_outcomes()

    @property
    def n_progs(self) -> int:
        """
        Return the number of programs

        :return: The number of programs with defined outcomes (usually this is a subset of all available programs)

        """

        return len(self.progs)

    def sample(self) -> None:
        """
        Perturb the values entered in the databook

        The :class:`Covout` instance is modified in-place. Note that the program outcomes are scalars
        that do not vary over time - therefore, :meth:`Covout.sample()` does not have a ``constant``
        argument.

        """

        if self.sigma is None:
            return

        for k, v in self.progs.items():
            self.progs[k] = v + self.sigma * np.random.randn(1)[0]
        # Perturb the interactions
        if self._interactions:
            for k, v in self.interactions.items():
                self.interactions[k] = v + self.sigma * np.random.randn(1)[0]
            tokens = ["%s=%.4f" % ("+".join(k), v) for k, v in self.interactions.items()]
            self.imp_interaction = ",".join(tokens)

        self.update_outcomes()

    def update_outcomes(self) -> None:
        """
        Update cache when outcomes change

        This method should be called whenever the baseline, program outcomes, or interaction outcomes change.
        It updates the internal cache so that get_outcome() uses the correct values. It's responsible for

        1. Sorting the programs by outcome value
        2. Compute the deltas relative to baseline
        3. Pre-compute the outcomes associated with every possible combination of programs

        """

        # First, sort the program dict by the magnitude of the outcome
        prog_tuple = [(k, v) for k, v in self.progs.items()]
        prog_tuple = sorted(prog_tuple, key=lambda x: -abs(x[1] - self.baseline))
        self._cached_progs = sc.odict()  # This list contains the perturbed/sampled values, in order
        for item in prog_tuple:
            self._cached_progs[item[0]] = item[1]
        self._deltas = np.array([x[1] - self.baseline for x in prog_tuple])  # Internally cache the deltas which are used

        # Precompute the combinations and associated modality interaction outcomes - it's computationally expensive otherwise
        # We need to store it in two forms
        # - An (ordered) vector of outcomes, which is used by additive and random to do the modality interaction in vectorized form
        # - A dict of outcomes, which is used by nested to look up the outcome using a tupled key of program indices
        combination_strings = [bin(x)[2:].rjust(self.n_progs, "0") for x in range(2**self.n_progs)]  # ['00','01','10',...]
        self.combinations = np.array([list(int(y) for y in x) for x in combination_strings])
        _combination_outcomes = []
        for prog_combination in self.combinations.astype(bool):
            _combination_outcomes.append(self.compute_impact_interaction(progs=prog_combination))
        self._combination_outcomes = np.array(_combination_outcomes)  # Reshape to column vector, since that's the shape of combination_coverage

    def __repr__(self):
        output = sc.prepr(self)
        output = sc.indent("   Parameter: ", self.par)
        output += sc.indent("  Population: ", self.pop)
        output += sc.indent("Baseline val: ", self.baseline)
        output += sc.indent("    Programs: ", ", ".join(["%s: %s" % (key, val) for key, val in self.progs.items()]))
        output += "\n"
        return output

    def get_outcome(self, prop_covered):
        """Return parameter value given program coverages

        The :class:`Covout` object contains a set of programs and outcomes. The :meth:`Covout.get_outcome` method
        returns the outcome value associated for coverage of each program. Don't forget that any given Covout instance
        is already specific to a ``(par,pop)`` combination

        :param prop_covered: A dict with ``{prog_name:coverage}`` containing at least all of the
                             programs in `self.progs`. Note that `coverage` is expected to be a ``np.array``
                             (such that that generated by :meth:`ProgramSet.get_prop_coverage`). However,
                             because the modality calculations only work for scalars, only the first entry
                             in the array will be used. Note that the coverage in this dictionary should be
                             dimensionless.
        :return: A scalar outcome (of type `np.double` or similar i.e. _not_ an array)

        """

        # Put coverages and deltas into array form
        outcome = self.baseline  # Accumulate the outcome by adding the deltas onto this

        if self.n_progs == 0:
            return outcome  # If there are no programs active, return the baseline value immediately
        elif self.n_progs == 1:
            return outcome + prop_covered[self._cached_progs.keys()[0]][0] * self._deltas[0]

        cov = []
        for prog in self._cached_progs.keys():
            cov.append(prop_covered[prog][0])
        cov = np.array(cov)

        # ADDITIVE CALCULATION
        if self.cov_interaction == "additive":
            # Outcome += c1*delta_out1 + c2*delta_out2

            # If sum(cov)<0 then there will be a divide by zero error. Also, need to divide by max(sum(cov),1) rather than sum(cov)
            # because otherwise, the coverages will be scaled UP to 1. So fastest just to check here
            if np.sum(cov) > 1:
                # Only keep the programs with nonzero coverage
                additive = np.maximum(cov - np.maximum(cov - (1 - (np.cumsum(cov) - cov)), 0), 0)
                remainder = 1 - additive
                random = cov - additive
                # If remainder is 0, then random must also be 0 i.e. it's always 0/0
                # This happens if the best program has coverage of exactly 1.0 which means it's entirely additive but also has no remainder
                random_portion = np.divide(random, remainder, out=np.zeros_like(random), where=remainder != 0)
                additive_portion_coverage = self.combinations * additive
                net_random = self.combinations * random_portion + (self.combinations ^ 1) * (1 - random_portion)
                combination_coverage = np.zeros((net_random.shape[0],))
                for i in range(0, net_random.shape[1]):
                    contribution = np.ones((net_random.shape[0],))
                    for j in range(0, net_random.shape[1]):
                        if i == j:
                            contribution *= additive_portion_coverage[:, j]
                        else:
                            contribution *= net_random[:, j]
                    combination_coverage += contribution
                outcome += np.sum(combination_coverage * self._combination_outcomes.ravel())
            else:
                outcome += np.sum(cov * self._deltas)

        # NESTED CALCULATION
        elif self.cov_interaction == "nested":
            # Outcome += c3*max(delta_out1,delta_out2,delta_out3) + (c2-c3)*max(delta_out1,delta_out2) + (c1 -c2)*delta_out1, where c3<c2<c1.
            idx = np.argsort(cov)
            prog_mask = np.full(cov.shape, fill_value=True)
            combination_coverage = np.zeros((self.combinations.shape[0],))

            for i in range(0, len(cov)):
                combination_index = int("0b" + "".join(["1" if x else "0" for x in prog_mask]), 2)
                if i == 0:
                    combination_coverage[combination_index] = cov[idx[i]]
                else:
                    combination_coverage[combination_index] = cov[idx[i]] - cov[idx[i - 1]]
                prog_mask[idx[i]] = False  # Disable this program at the next iteration

            outcome += np.sum(combination_coverage * self._combination_outcomes.ravel())

        # RANDOM CALCULATION
        elif self.cov_interaction == "random":
            # Outcome += c1(1-c2)* delta_out1 + c2(1-c1)*delta_out2 + c1c2* max(delta_out1,delta_out2)
            combination_coverage = np.prod(self.combinations * cov + (self.combinations ^ 1) * (1 - cov), axis=1)
            outcome += np.sum(combination_coverage.ravel() * self._combination_outcomes.ravel())
        else:
            raise Exception('Unknown reachability type "%s"', self.cov_interaction)

        return outcome

    def compute_impact_interaction(self, progs: np.array) -> float:
        """
        Return the output for a given combination of programs

        The outcome for various combinations of programs is cached prior to running the model.
        This function retrieves the appropriate delta given a boolean array flagging which
        programs are active

        :param progs: A numpy boolean array, with length equal to the number of programs
        :return: The delta value corresponding to the specified combination of programs

        """

        if not any(progs):
            return 0.0
        else:
            progs_active = frozenset(np.array(self._cached_progs.keys())[progs])

        if progs_active in self._interactions:
            # If the combination of programs has an explicitly specified outcome, then use it
            return self._interactions[progs_active]
        elif self.imp_interaction is not None and self.imp_interaction.lower() == "synergistic":
            raise NotImplementedError
        else:
            # Otherwise, do the 'best' interaction and return the delta with the largest magnitude
            tmp = self._deltas[progs]
            idx = np.argmax(abs(tmp))
            return tmp[idx]



import atomica as at
import sciris as sc
import os
import pytest
import sys
import itertools

# # P = at.Project(framework='dummy_framework.xlsx',databook='dummy_databook.xlsx')
# # # P.run_sim()
# # P = at.demo('tb',do_run=True)
# # P.results[0].plot()
# # P = at.Project.load('asdf2.prj')
#
# # P = at.Project(framework=at.LIBRARY_PATH+'tb_framework.xlsx')
# # P.load_databook(at.LIBRARY_PATH+'tb_databook.xlsx')
# P = at.demo('tb',do_run=True)

testdir = at.parent_dir()
tmpdir = testdir / "temp"


def test_timed_initialization():

    F = at.ProjectFramework(testdir / "timed_initialization_test_framework.xlsx")
    D = at.ProjectData.new(F, [2018], 1, 0)
    D.tdve["tx_rate"].ts[0].insert(2018, 0)
    D.tdve["tx_rate"].ts[0].insert(2021, 0)
    D.tdve["tx_rate"].ts[0].insert(2022, 0.5)

    def set_initialization_basic(F, D, year, y_factor=True):
        # Conventional approach using Y-factor adjustment only
        res = P.run_sim(P.parsets[0].make_constant(2018))
        ps2 = at.ParameterSet(F, D, "basic")

        # Set initial compartment sizes for initialization quantities
        for qty in itertools.chain(F.characs.index[F.characs["setup weight"] > 0], F.comps.index[F.comps["setup weight"] > 0]):
            for pop in D.pops:
                if y_factor:
                    ps2.pars[qty].meta_y_factor = 1
                    ps2.pars[qty].y_factor[pop] = res.get_variable(qty, pop)[0].vals[-1] / ps2.pars[qty].ts[pop].interpolate(year)
                else:
                    ps2.pars[qty].ts[pop].insert(year, res.get_variable(qty, pop)[0].vals[-1])

        return ps2

    P = at.Project(framework=F, databook=D, do_run=False)
    P.settings.sim_dt = 1 / 12
    P.settings.sim_start = 2018
    P.settings.sim_end = 2023

    res1 = P.run_sim()

    # Basic steady state initialization
    ps = set_initialization_basic(F, D, 2018)
    res2 = P.run_sim(ps, result_name="basic")

    # Advanced steady state initialization
    ps2 = at.ParameterSet(F, D)
    ps2.set_initialization(res1, 2021)
    res3 = P.run_sim(ps2, result_name="advanced")

    ps2.pars["alive"].y_factor[0] = 1.1

    ps2.save_calibration(tmpdir / "test_cal.xlsx")
    res3 = P.run_sim(ps2, result_name="advanced")

    ps4 = P.make_parset("saved")
    ps4.load_calibration(tmpdir / "test_cal.xlsx")
    res4 = P.run_sim(ps4, result_name="loaded")

    d = at.PlotData([res1, res2, res3, res4], ["sus", "inf", "rec"])
    at.plot_series(d)


if __name__ == "__main__":
    test_timed_initialization()



import numpy as np
import atomica as at
import os
import matplotlib.pyplot as plt


def test_stochastic():

    testdir = at.parent_dir()  # Must be relative to current file to work with tox

    F0 = at.ProjectFramework(testdir / "framework_sir_dynamic.xlsx")
    P0 = at.Project(name="test", framework=F0, do_run=False)
    P0.load_databook(databook_path=testdir / "databook_sir_dynamic.xlsx", make_default_parset=True, do_run=False)
    baseline = P0.run_sim(result_name="default")

    F = at.ProjectFramework(testdir / "framework_stochastic_test.xlsx")
    P = at.Project(name="test", framework=F, do_run=False)
    P.load_databook(databook_path=testdir / "databook_sir_dynamic.xlsx", make_default_parset=True, do_run=False)
    ensemble = at.Ensemble(mapping_function=lambda x: at.PlotData(x), baseline_results=baseline)
    ensemble.run_sims(P, "default", n_samples=100, result_names="default")
    fig = ensemble.plot_series()
    at.relabel_legend(fig, ["Susceptible", "Infected", "Recovered"])
    plt.ylabel("Number of people")


if __name__ == "__main__":
    test_stochastic()



"""
Implements the :class:`Project` user interface for Atomica

The :class:`Project` class serves as the primary user interface for
Atomica. Almost all functionality can be accessed via this interface.

A project is based around 5 major lists:
    1. parsets -- an odict of parameter sets
    2. progsets -- an odict of program sets
    3. scens -- an odict of scenario structures
    4. optims -- an odict of optimization structures
    5. results -- an odict of results associated with parsets, scens, and optims

In addition, a project contains:
    1. data -- loaded from the spreadsheet
    2. settings -- timestep, indices, etc.
    3. various kinds of metadata -- project name, creation date, etc.

"""

from .version import version, gitinfo
from .calibration import calibrate
from .data import ProjectData
from .framework import ProjectFramework
from .model import run_model
from .parameters import ParameterSet

from .programs import ProgramSet
from .scenarios import Scenario, ParameterScenario, CombinedScenario, BudgetScenario, CoverageScenario
from .optimization import Optimization, optimize, InvalidInitialConditions
from .system import logger
from .utils import NDict, evaluate_plot_string, NamedItem, parallel_progress, Quiet
from .plotting import PlotData, plot_series
from .results import Result
from .migration import migrate
import sciris as sc
import numpy as np
import tqdm
import logging
from datetime import timezone
import functools

__all__ = ["ProjectSettings", "Project"]


class ProjectSettings:
    def __init__(self, sim_start=2000, sim_end=2035, sim_dt=0.25):
        self._sim_start = sim_start
        self._sim_dt = sim_dt
        self._sim_end = 0.0
        self.update_time_vector(end=sim_end)

    def __repr__(self):
        """Print object"""
        output = sc.prepr(self)
        return output

    @property
    def sim_start(self):
        return self._sim_start

    @property
    def sim_end(self):
        return self._sim_end

    @property
    def sim_dt(self):
        return self._sim_dt

    @sim_start.setter
    def sim_start(self, sim_start):
        self._sim_start = sim_start

    @sim_end.setter
    def sim_end(self, sim_end):
        self._sim_end = self.sim_start + np.ceil((sim_end - self.sim_start) / self.sim_dt) * self.sim_dt
        if sim_end != self._sim_end:
            logger.info(f"Changing sim end from {sim_end} to {self._sim_end} ({(self._sim_end - self._sim_start) / self._sim_dt:.0f} timesteps)")

    @sim_dt.setter
    def sim_dt(self, sim_dt):
        assert sim_dt > 0, "Simulation timestep must be greater than 0"
        self._sim_dt = sim_dt
        self.sim_end = self.sim_end  # Call the setter function to change sim_end if it is no longer valid

    @property
    def tvec(self) -> np.ndarray:
        """
        Return simulation time vector

        This method uses `arange` with a step size of 1, then multiplies with the correct step size and add sim start time.

        :return: Array of simulation times

        """

        return np.arange(round((self.sim_end - self.sim_start) / self.sim_dt) + 1) * self.sim_dt + self.sim_start

    def update_time_vector(self, start: float = None, end: float = None, dt: float = None) -> None:
        """
        Update the project simulation times

        :param start: Optionally provide new start year (e.g. '2018')
        :param end: Optionally provide new end year (e.g. '2035')
        :param dt: Optionally provide new step size, in years (e.g. '0.25' for quarterly steps

        """

        if start is not None:
            self.sim_start = start

        if dt is not None:
            if end is None:
                self.sim_dt = dt  # Update the simulation end year automatically to account for the new timestep
            else:
                self._sim_dt = dt  # Change sim_dt internally so that the sim_end is only changed once

        if end is not None:
            self.sim_end = end


class Project(NamedItem):
    """
    Main simulation container

    A Project provides a centralized point of storage when working with Atomica. It contains

    - A framework
    - Data
    - Parameters
    - Programs
    - Scenarios and optimizations
    - Results

    Importantly, it is generally assumed that saving and loading work is done by saving and
    loading projects.

    :param name:
    :param framework: a Framework to use. This could be
                      - A filename to an Excel file on disk
                      - An sc.Spreadsheet instance
                      - A ProjectFramework instance
                      - None (this should generally not be used though!)
    :param databook: The path to a databook file. The databook will be loaded into Project.data and the spreadsheet saved to Project.databook
    :param do_run: If True, a simulation will be run upon project construction
    :param sim_start: Optionally specify the sim start year (otherwise will default to the first year in the databook)
    :param sim_end: Optionally specify simulation end year
    :param sim_dt: Optionally specify simulation timestep
    """

    def __init__(self, name="default", framework=None, databook=None, do_run: bool = True, sim_start: float = None, sim_end: float = None, sim_dt: float = None):
        NamedItem.__init__(self, name)

        if isinstance(framework, ProjectFramework):
            self.framework = framework
        elif framework is None:
            logger.warning("Project was constructed without a Framework - a framework should be provided")
            self.framework = None
        else:
            self.framework = ProjectFramework(inputs=framework)

        # Define the structure sets
        self.parsets = NDict()  #: Dictionary of :class:`ParameterSet` instances
        self.progsets = NDict()
        self.scens = NDict()
        self.optims = NDict()
        self.results = NDict()

        # Define metadata
        self.uid = sc.uuid()
        self.version = version
        self.gitinfo = sc.dcp(gitinfo)
        self.filename = None

        self.progbook = None  # This will contain an sc.Spreadsheet when the user loads one
        self.settings = ProjectSettings()  # Global settings

        self._update_required = False  # This flag is set to True by migration is the result objects contained in this Project are out of date due to a migration change

        # Load project data, if available
        if framework and databook:
            self.load_databook(databook_path=databook, make_default_parset=True, do_run=False)
        elif databook:
            logger.warning("Project was constructed without a Framework - databook spreadsheet is being saved to project, but data is not being loaded")
            self.databook = sc.Spreadsheet(databook)  # Just load the spreadsheet in so that it isn't lost
            self.data = None
        else:
            self.databook = None  # This will contain an sc.Spreadsheet when the user loads one
            self.data = None

        # If the user has overridden any of these, apply those values after the databook is loaded
        # Otherwise, the default logic in `Project.load_databook()` will be used
        self.settings.update_time_vector(start=sim_start, end=sim_end, dt=sim_dt)

        # Run the model, but only if both a framework and databook were provided
        if framework and databook and do_run:
            self.run_sim(parset="default", store_results=True)

    def __repr__(self):
        """Print out useful information when called"""
        output = sc.objrepr(self)
        output += "      Project name: %s\n" % self.name
        output += "    Framework name: %s\n" % self.framework.name
        output += "\n"
        output += "    Parameter sets: %i\n" % len(self.parsets)
        output += "      Program sets: %i\n" % len(self.progsets)
        output += "         Scenarios: %i\n" % len(self.scens)
        output += "     Optimizations: %i\n" % len(self.optims)
        output += "      Results sets: %i\n" % len(self.results)
        output += "\n"
        output += "   Atomica version: %s\n" % self.version
        output += "      Date created: %s\n" % sc.getdate(self.created.replace(tzinfo=timezone.utc).astimezone(tz=None), dateformat="%Y-%b-%d %H:%M:%S %Z")
        output += "     Date modified: %s\n" % sc.getdate(self.modified.replace(tzinfo=timezone.utc).astimezone(tz=None), dateformat="%Y-%b-%d %H:%M:%S %Z")
        #        output += '  Datasheet loaded: %s\n' % sc.getdate(self.databookloaddate)
        output += "        Git branch: %s\n" % self.gitinfo["branch"]
        output += "          Git hash: %s\n" % self.gitinfo["hash"]
        output += "               UID: %s\n" % self.uid
        output += "============================================================\n"
        return output

    # @property
    # def pop_names(self):
    #     if self.data is None:
    #         return []
    #     else:
    #         return list(self.data.pops.keys())
    #
    # @property
    # def pop_labels(self):
    #     if self.data is None:
    #         return []
    #     else:
    #         return list([x['label'] for x in self.data.pops.values()])

    #######################################################################################################
    # Methods for I/O and spreadsheet loading
    #######################################################################################################
    def create_databook(self, databook_path=None, num_pops=1, num_transfers=0, data_start=2000.0, data_end=2020.0, data_dt=1.0):
        """Generate an empty data-input Excel spreadsheet corresponding to the framework of this project."""
        if databook_path is None:
            databook_path = "./databook_" + self.name + ".xlsx"
        data = ProjectData.new(self.framework, np.arange(data_start, data_end + data_dt, data_dt), pops=num_pops, transfers=num_transfers)
        data.save(databook_path)
        return data

    def load_databook(self, databook_path, make_default_parset: bool = True, do_run: bool = True) -> None:
        """
        Load a data spreadsheet

        :param databook_path: Databook input. Supported inputs are:
                                - A path string, which will load a file from disk
                                - An `sc.Spreadsheet` containing the contents of a databook
                                - A `ProjectData` instance
        :param make_default_parset: If True, a Parset called "default" will be immediately created from the newly-added data
        :param do_run: If True, a simulation will be run using the new parset

        """

        if isinstance(databook_path, ProjectData):
            # If the user passed in a ProjectData instance, use it directly
            data = databook_path
            databook = None
        else:
            if isinstance(databook_path, sc.Spreadsheet):
                databook = databook_path
            else:
                databook = sc.Spreadsheet(databook_path)
            data = ProjectData.from_spreadsheet(databook, self.framework)

        # If there are existing progsets, make sure the new data is consistent with them
        if self.progsets:
            data_pops = set((x, y["label"]) for x, y in data.pops.items())
            for progset in self.progsets.values():
                assert data_pops == set((x, y["label"]) for x, y in progset.pops.items()), "Existing progsets exist with populations that do not match the new databook"

        self.data = data
        self.data.validate(self.framework)  # Make sure the data is suitable for use in the Project (as opposed to just manipulating the databook)
        self.databook = sc.dcp(databook)
        self.modified = sc.now(utc=True)
        self.settings.update_time_vector(start=self.data.start_year)  # Align sim start year with data start year.

        if not (self.framework.comps["is source"] == "y").any():
            self.settings.update_time_vector(end=self.data.end_year + 5.0)  # Project only forecasts 5 years if not dynamic (with births)

        if make_default_parset:
            self.make_parset(name="default")
        if do_run:
            if not make_default_parset:
                logger.warning("Project has been requested to run a simulation after loading databook, " "despite no request to create a default parameter set.")
            self.run_sim(parset="default", store_results=True)

    def make_parset(self, name="default"):
        """Transform project data into a set of parameters that can be used in model simulations."""
        self.parsets.append(ParameterSet(self.framework, self.data, name))
        return self.parsets[name]

    def make_progbook(self, progbook_path: str = None, progs=None, data_start: float = None, data_end: float = None) -> str:
        """
        Save a blank program book

        This method will create a temporary ``ProgramSet`` and save it to the requested path

        :param progbook_path: Path to save program book
        :param progs: A program specification supposed by ``ProgramSet.new()``
        :param data_start: The start year for data entry
        :param data_end: The end year for data entry
        :return: The full path of the newly created program book

        """

        # Get filepath
        if self.data is None:
            errormsg = "Please upload a databook before creating a program book. The databook defines which populations will appear in the program book."
            raise Exception(errormsg)
        full_path = sc.makefilepath(filename=progbook_path, default=self.name, ext="xlsx", makedirs=False)  # Directory will be created later in progset.save()
        if data_start is None:
            data_start = self.data.tvec[0]
        if data_end is None:
            data_end = self.data.tvec[-1]
        progset = ProgramSet.new(tvec=np.arange(data_start, data_end + 1), progs=progs, framework=self.framework, data=self.data)
        progset.save(full_path)
        return full_path

    def load_progbook(self, progbook_path=None, name="default"):
        """
        Create a ProgramSet given a progbook

        :param progbook_path: Path to a program spreadsheet or an AtomicaSpreadsheet instance
        :param name: The name to assign to the new ProgramSet
        :return: The newly created ProgramSet (also stored in ``self.progsets``)

        """

        if self.data is None:
            errormsg = "Please upload a databook before uploading a program book. The databook contains the population definitions required to read the program book."
            raise Exception(errormsg)

        if sc.isstring(progbook_path):
            full_path = sc.makefilepath(filename=progbook_path, default=self.name, ext="xlsx", makedirs=False)
            progbook_spreadsheet = sc.Spreadsheet(full_path)
        else:
            progbook_spreadsheet = progbook_path

        progset = ProgramSet.from_spreadsheet(spreadsheet=progbook_spreadsheet, framework=self.framework, data=self.data, name=name)
        progset.validate()
        self.progbook = sc.dcp(progbook_spreadsheet)  # Actually a shallow copy is fine here because AtomicaSpreadsheet contains no mutable properties
        self.progsets.append(progset)
        return progset

    def make_scenario(self, which: str = "combined", **kwargs) -> Scenario:
        """
        Make new scenario and store in Project

        :param which: String identifying type - one of ``['parameter','budget','coverage','combined']``
        :param kwargs: Arguments to pass to appropriate :class:`Scenario` constructor
        :return: New scenario instance

        """

        types = {"parameter": ParameterScenario, "budget": BudgetScenario, "coverage": CoverageScenario, "combined": CombinedScenario}
        scenario = types[which](**kwargs)
        self.scens.append(scenario)
        return scenario

    #    #######################################################################################################
    #    ### Utilities
    #    #######################################################################################################

    def parset(self, key=None, verbose=2):
        """Shortcut for getting a parset"""
        if key is None:
            key = -1
        if isinstance(key, ParameterSet):
            return key  # It's already a parameter set, do nothing
        else:
            try:
                return self.parsets[key]
            except Exception:
                sc.printv('Warning, parset "%s" not found!' % key, 1, verbose)
                return None

    def progset(self, key=None, verbose=2):
        """Shortcut for getting a progset"""
        if key is None:
            key = -1
        if isinstance(key, ProgramSet):
            return key  # It's already a program set, do nothing
        else:
            try:
                return self.progsets[key]
            except Exception:
                sc.printv('Warning, progset "%s" not found!' % key, 1, verbose)
                return None

    def scen(self, key=None, verbose=2):
        """Shortcut for getting a scenario"""
        if key is None:
            key = -1
        if isinstance(key, Scenario):
            return key  # It's already a scenario, do nothing
        else:
            try:
                return self.scens[key]
            except Exception:
                sc.printv('Warning, scenario "%s" not found!' % key, 1, verbose)
                return None

    def optim(self, key=None, verbose=2):
        """Shortcut for getting an optimization"""
        if key is None:
            key = -1
        if isinstance(key, Optimization):
            return key  # It's already an optimization, do nothing
        else:
            try:
                return self.optims[key]
            except Exception:
                sc.printv('Warning, scenario "%s" not found!' % key, 1, verbose)
                return None

    def result(self, key=None, verbose=2):
        """Shortcut for getting an result -- a little special since they don't have a fixed type"""
        if key is None:
            key = -1
        if not sc.isstring(key) and not sc.isnumber(key) and not isinstance(key, tuple):
            if not isinstance(key, [Result, list, sc.odict]):
                print('Warning: result "%s" is of unexpected type: "%s"' % (key, type(key)))
            return key  # It's not something that looks like a key
        else:
            try:
                return self.scens[key]
            except Exception:
                sc.printv('Warning, scenario "%s" not found!' % key, 1, verbose)
                return None

        if key is None:
            key = -1
        try:
            return self.results[key]
        except Exception:
            return sc.printv('Warning, results "%s" not found!' % key, 1, verbose)  # Returns None

    #######################################################################################################
    # Methods to perform major tasks
    #######################################################################################################

    def plot(self, results=None, key=None, outputs=None, pops=None):
        def get_supported_plots():
            df = self.framework.sheets["plots"][0]
            plots = sc.odict()
            for name, output in zip(df["name"], df["quantities"]):
                plots[name] = evaluate_plot_string(output)
            return plots

        if outputs is None:
            supported_plots = get_supported_plots()
            outputs = [{plot_name: supported_plots[plot_name]} for plot_name in supported_plots.keys()]
        if results is None:
            results = self.result(key)

        allfigs = []
        for output in outputs:
            try:
                if not isinstance(list(output.values())[0], list):
                    output = list(output.values())[0]
                plotdata = PlotData(results, outputs=output, project=self, pops=pops)
                figs = plot_series(plotdata, axis="pops", plot_type="stacked", legend_mode="together")
                allfigs += figs
            except Exception as e:
                print("WARNING, %s failed (%s)" % (output, str(e)))
        return allfigs

    def update_settings(self, sim_start=None, sim_end=None, sim_dt=None):
        """Modify the project settings, e.g. the simulation time vector."""
        self.settings.update_time_vector(start=sim_start, end=sim_end, dt=sim_dt)

    def run_sim(self, parset=None, progset=None, progset_instructions=None, store_results=False, result_name: str = None):
        """
        Run a single simulation

        This function is the main entry point for running model simulations, given a
        parset and optionally program set + instructions.

        :param parset: A :class:`ParameterSet` instance, or the name of a parset contained in ``self.parsets``.
                        If ``None``, then the most recently added parset will be used (the last entry in ``self.parsets``)
        :param progset: Optionally pass in a :class:`ProgramSet` instance, or the name of a progset contained in ``self.progsets``
        :param progset_instructions: A :class:`ProgramInstructions` instance. Programs will only be used if a instructions are provided
        :param store_results: If True, then the result will automatically be stored in ``self.results``
        :param result_name: Optionally assign a specific name to the result (otherwise, a unique default name will automatically be selected)
        :return: A :class:`Result` instance

        """

        parset = self.parset(parset)
        if progset is not None:
            progset = self.progset(progset)
            if progset_instructions is None:
                logger.info("Program set '%s' will be ignored while running project '%s' due to the absence of program set instructions", progset.name, self.name)

        if result_name is None:
            base_name = "parset_" + parset.name
            if progset is not None:
                base_name = base_name + "_progset_" + progset.name
            k = 1
            result_name = base_name
            while result_name in self.results:
                result_name = base_name + "_" + str(k)
                k += 1

        tm = sc.tic()
        result = run_model(settings=self.settings, framework=self.framework, parset=parset, progset=progset, program_instructions=progset_instructions, name=result_name)
        logger.info('Elapsed time for running "%s": %ss', self.name, sc.sigfig(sc.toc(tm, output=True), 3))
        if store_results:
            self.results.append(result)

        return result

    def run_sampled_sims(self, parset, progset=None, progset_instructions=None, result_names=None, n_samples: int = 1, parallel=False, max_attempts=None, num_workers=None) -> list:
        """
        Run sampled simulations

        This method samples from the parset (and progset if provided). It is separate from `run_sim` for
        several reasons

        - To avoid inadvertantly blowing up the size of the project, `run_sampled_sims` does not support automatic result saving
        - `run_sim` always returns a `Result` - if rolled into one function, the return type would not be predictable
        - `run_sim` only takes in a single ``ProgramInstructions`` and ``result_name`` whereas `run_sampled_sims` supports iteration
           over multiple instructions

        The other common scenario is having multiple results

        :param n_samples: An integer number of samples
        :param parset: A :class:`ParameterSet` instance
        :param progset: Optionally a :class:`ProgramSet` instance
        :param progset_instructions: This can be a list of instructions
        :param result_names: Optionally specify names for each result. The most common usage would be when passing in a list of program instructions
                             corresponding to different budget scenarios. The result names should be a list the same length as the instructions, or
                             containing a single element if not using programs.
        :param parallel: If True, run simulations in parallel (on Windows, must have ``if __name__ == '__main__'`` gating the calling code)
        :param max_attempts: Number of retry attempts for bad initializations
        :param num_workers: If ``parallel`` is True, this determines the number of parallel workers to use (default is usually number of CPUs)
        :return: A list of Results that can be passed to `Ensemble.update()`. If multiple instructions are provided, the return value of this
                 function will be a list of lists, where the inner list iterates over different instructions for the same parset/progset samples.
                 It is expected in that case that the Ensemble's mapping function would take in a list of results

        """

        assert (not progset) == (not progset_instructions), "If running with programs, both a progset and instructions must be provided"

        parset = self.parset(parset)
        progset = self.progset(progset) if progset is not None else None
        progset_instructions = sc.promotetolist(progset_instructions, keepnone=True)

        if not result_names:
            if len(progset_instructions) > 1:
                result_names = ["instructions_%d" % (i) for i in range(len(progset_instructions))]
            else:
                result_names = ["default"]
        else:
            result_names = sc.promotetolist(result_names)
            assert (len(result_names) == 1 and not progset) or (len(progset_instructions) == len(result_names)), "Number of result names must match number of instructions"

        show_progress = n_samples > 1 and logger.getEffectiveLevel() <= logging.INFO

        if parallel:
            fcn = functools.partial(_run_sampled_sim, proj=self, parset=parset, progset=progset, progset_instructions=progset_instructions, result_names=result_names, max_attempts=max_attempts)
            results = parallel_progress(fcn, n_samples, show_progress=show_progress, num_workers=num_workers)
        elif show_progress:
            # Print the progress bar if the logging level was INFO or lower
            # This means that the user can still set the logging level higher e.g. WARNING to suppress output from Atomica in general
            # (including any progress bars)
            with Quiet():
                results = [_run_sampled_sim(self, parset, progset, progset_instructions, result_names, max_attempts=max_attempts) for _ in tqdm.trange(n_samples)]
        else:
            results = [_run_sampled_sim(self, parset, progset, progset_instructions, result_names, max_attempts=max_attempts) for _ in range(n_samples)]

        return results

    def calibrate(self, parset=None, adjustables=None, measurables=None, save_to_project=False, new_name=None, default_min_scale=0.0, default_max_scale=2.0, default_weight=1.0, default_metric="fractional", yaml=None, **kwargs) -> ParameterSet:
        """
        Method to perform automatic calibration.

        The adjustables argument should be a list in the form of...
            [par_name_1, par_name_2, charac_name_1]
        ...or...
            [(par_name_1, pop_1, min_scale_1, max_scale_1)
             (par_name_2, None, min_scale_2, max_scale_2),
             (charac_name_1, pop_2, min_scale_3, max_scale_3)]
        The former instructs specified parameter values for all populations to be varied between default scaling limits.
        The latter varies specified parameters for specified populations, within specified scaling limits.
        'None' in the population position represents independent scaling across all populations.

        The measurables argument should be a list in the form of...
            [charac_name_1, charac_name_2]
        ...or...
            [(charac_name_1, pop_1, weight_1, "fractional")
             (charac_name_2, None, weight_2, "wape")]
        The former calculates a 'fractional' data comparison metric across specified characteristics for all pops.
        The latter calculates its metric for specified populations and for both specified weights and metric types.
        'None' represents combining the metric across all populations.

        To calibrate a project-attached parameter set in place, provide its key as the new name argument to this method.
        Current fitting metrics are: "fractional", "meansquare", "wape"
        Note that scaling limits are absolute, not relative.

        Additional keyword arguments will be passed to `at.calibrate` so additional arguments such as `max_time`
        or `methods` can be provided to this function.

        :param parset:
        :param adjustables:
        :param measurables:
        :param save_to_project:
        :param new_name:
        :param default_min_scale:
        :param default_max_scale:
        :param default_weight:
        :param default_metric:
        :param yaml: Optionally specify a YAML file or YAML calibration dictionary to use for calibration
        :param kwargs:
        :return:

        """

        if parset is None:
            parset = self.parsets[-1]
        elif not isinstance(parset, ParameterSet):
            parset = self.parsets[parset]

        if yaml is not None:
            import atomica.yaml_calibration  # Avoid circular import

            assert adjustables is None, "If a YAML file is specified, adjustables should not be set"
            assert measurables is None, "If a YAML file is specified, measurables should not be set"
            assert "time_period" not in kwargs, "If a YAML file is specified, time_period should not be set - instead, set cal_start and cal_end in the YAML file"
            new_parset = atomica.yaml_calibration.run(yaml, self, parset, **kwargs)
        else:
            if adjustables is None:
                adjustables = list(self.framework.pars.index[~self.framework.pars["calibrate"].isnull()])
                adjustables += list(self.framework.comps.index[~self.framework.comps["calibrate"].isnull()])
                adjustables += list(self.framework.characs.index[~self.framework.characs["calibrate"].isnull()])

            if measurables is None:
                measurables = list(self.framework.comps.index)
                measurables += list(self.framework.characs.index)

            for index, adjustable in enumerate(adjustables):
                if sc.isstring(adjustable):  # Assume that a parameter name was passed in if not a tuple.
                    adjustables[index] = (adjustable, None, default_min_scale, default_max_scale)

            for index, measurable in enumerate(measurables):
                if sc.isstring(measurable):  # Assume that a parameter name was passed in if not a tuple.
                    measurables[index] = (measurable, None, default_weight, default_metric)

            new_parset = calibrate(project=self, parset=parset, pars_to_adjust=adjustables, output_quantities=measurables, **kwargs)

        if new_name is None:
            new_name = parset.name + " (auto-calibrated)"
        new_parset.name = new_name  # The new parset is a calibrated copy of the old, so change id.

        if save_to_project:
            self.parsets.append(new_parset)

        return new_parset

    def run_scenarios(self, store_results: bool = True) -> list:
        """
        Run all active scenarios

        :param store_results: If True, results will be appended to the project
        :return: List of results (one for each active scenario)

        """

        results = []
        for scenario in self.scens.values():
            if scenario.active:
                result = scenario.run(project=self, store_results=store_results)
                results.append(result)
        return results

    def run_optimization(self, optimname=None, maxtime=None, maxiters=None, store_results=True):
        """Run an optimization"""
        optim_ins = self.optim(optimname)
        optim, unoptimized_instructions = optim_ins.make(project=self)
        if maxtime is not None:
            optim.maxtime = maxtime
        if maxiters is not None:
            optim.maxiters = maxiters
        parset = self.parset(optim.parsetname)
        progset = self.progset(optim.progsetname)
        original_end = self.settings.sim_end
        self.settings.sim_end = optim_ins.json["end_year"]  # Simulation should be run up to the user's end year
        try:
            optimized_instructions = optimize(self, optim, parset, progset, unoptimized_instructions)
        except InvalidInitialConditions:
            if optim_ins.json["optim_type"] == "money":
                raise Exception("It was not possible to achieve the optimization target even with an increased budget. Specify or raise upper limits for spending, or decrease the optimization target")
            else:
                raise  # Just raise it as-is

        self.settings.sim_end = original_end  # Note that if the end year is after the original simulation year, the result won't be visible (although it will have been optimized for)
        optimized_result = self.run_sim(parset=parset, progset=progset, progset_instructions=optimized_instructions, result_name="Optimized", store_results=store_results)
        unoptimized_result = self.run_sim(parset=parset, progset=progset, progset_instructions=unoptimized_instructions, result_name="Baseline", store_results=store_results)
        results = [unoptimized_result, optimized_result]
        return results

    def save(self, filename: str = None, folder: str = None) -> str:
        """
        Save binary project file

        This method saves the entire project as a binary blob to disk

        :param filename: Name of the file to save
        :param folder: Optionally specify a folder
        :return: The full path of the file that was saved

        """

        fullpath = sc.makefilepath(filename=filename, folder=folder, default=[self.filename, self.name], ext="prj", sanitize=True)
        self.filename = fullpath
        sc.saveobj(fullpath, self)
        return fullpath

    @staticmethod
    def load(filepath):
        """
        Load binary project file

        This method is an alternate constructor that is used to load a binary file
        saved using :meth:`Project.save`. Migration is automatically performed as
        part of the loading operation.

        :param filepath: The file path/name to load
        :return: A new :class:`Project` instance

        """

        P = sc.loadobj(filepath, die=True)
        assert isinstance(P, Project)
        return P

    def __setstate__(self, d):
        self.__dict__ = d
        P = migrate(self)
        self.__dict__ = P.__dict__


def _run_sampled_sim(proj, parset, progset, progset_instructions: list, result_names: list, max_attempts: int = None):
    """
    Internal function to run simulation with sampling

    This function is intended for internal use only. It's purpose is to facilitate the implementation
    of parallelization. It should normally be called via :meth:`Project.run_sim`.

    This standalone function samples and runs a simulation. It is a standalone function rather than
    a method of :class:`Project` or :class:`Ensemble` so that it can be pickled for use in
    ``sc.parallelize`` (otherwise, an error relating to not being able to pickle local functions or the
    base class gets raised).

    A sampled simulation may result in bad initial conditions. If that occurs, the parameters and program
    set will be resampled up to a maximum of ``n_attempts`` times, after which an error will be raised.

    :param proj: A :class:`Project` instance
    :param parset: A :class:`ParameterSet` instance
    :param progset: A :class:`ProgramSet` instance
    :param progset_instructions: A list of instructions to run against a single sample
    :param result_names: A list of result names (strings)
    :param max_attempts: Maximum number of sampling attempts before raising an error
    :return: A list of results that either contains 1 result, or the same number of results as instructions

    """

    from .model import BadInitialization  # avoid circular import

    if max_attempts is None:
        max_attempts = 50

    attempts = 0
    while attempts < max_attempts:
        try:
            if progset:
                sampled_parset = parset.sample()
                sampled_progset = progset.sample()
                results = [proj.run_sim(parset=sampled_parset, progset=sampled_progset, progset_instructions=x, result_name=y) for x, y in zip(progset_instructions, result_names)]
            else:
                sampled_parset = parset.sample()
                results = [proj.run_sim(parset=sampled_parset, result_name=y) for y in result_names]
            return results
        except BadInitialization:
            attempts += 1
    raise Exception("Failed simulation after %d attempts - something might have gone wrong" % (max_attempts))
